(define *SchemeVersion* "Gambit 2.2")


;;; The following macro is borrowed from meroonV3 by Christian Queinnec.
;;; It is a macro definer that allows macros to
;;; remain in the compiled image. 

(##define-macro (defmacro call . body)
  `(begin (eval '(##define-macro ,call . ,body))
          (##define-macro ,call . ,body) ) )

(eval '(##define-macro (defmacro call . body)
  `(begin (eval '(##define-macro ,call . ,body))
          (##define-macro ,call . ,body) ) ))


 ;; some general new constructs
  
(defmacro (unless expression . body) `(if ,expression #f (begin ,@body)))
(defmacro (when expression . body) `(if ,expression (begin ,@body) #f))
(defmacro (1- x) `(- ,x 1))
(defmacro (1+ x) `(+ ,x 1))
(defmacro (<- x k)`(vector-ref ,x ,k))
(defmacro (-> x k y)`(vector-set! ,x ,k ,y))
;;============================================================================
;; Defining Blocks and Exiting from them.
;;----------------------------------------------------------------------------
(defmacro (block name . code)
   `(call/cc (lambda (,name) ,@code)))
(defmacro (return-from name . arg) 
                `(,name ,(if (null? arg) `#t (car arg))))

;; Syntactic sugar for multiple value return and let
;; (multiple-let ((x y ...)(function with multiple values)) <forms>) 
(defmacro (multiple-let assignment . body)
   (let
     ((list-vars (car assignment))
      (multiple-function (cadr assignment))
      (var (gensym)))
     (define (inner)
       (let((i -1))
         (lambda(x)
           (set! i (+ 1 i))
           (cons x (list `(list-ref ,var ,i))))))
       
     `(let*
        ((,var ,multiple-function)
         ,@(map (inner) list-vars))
        ,@body)))

(defmacro (return-list . lista) `(list ,@lista))

(defmacro (mpush x stack)
                `(set! ,stack (cons ,x ,stack)))

(defmacro (mpop stack)
                `(let ((x0 (car ,stack))) (set! ,stack (cdr ,stack)) x0))

(defmacro (for var lower upper . body)
  (let ((loop (gensym)))
   `(let ,loop ((,var ,lower))
      (if (<= ,var ,upper) (begin ,@body (,loop (+ 1 ,var)))))))

(defmacro (while pair . body)
  (let ((label (gensym))
        (var (car pair))
        (lista (cadr pair)))
    `(let ,label ((,var ,lista))
       (cond ((pair? ,var) 
              ,@body
              (,label (cdr ,var)))))))

;;----------------------useful functions-----------------------------------
(define (make-counter)
  (let ((n 0))
    (lambda()(let ((x n))(set! n (+ 1 n)) x))))

(define (f-or . lista)
  (if (pair? lista)
      (or (car lista)(apply f-or (cdr lista)))
      #f))

(define (f-and . lista)
  (if (pair? lista)
    (and (car lista)(apply f-and (cdr lista)))
    #t))
  
(define (nconc a b)
  (define (inner-conc x)
    (cond ((null? (cdr x)) (set-cdr! x b))
          (else (inner-conc (cdr x)))))
  (if (pair? a) 
    (begin
      (inner-conc a)
      a)
    #f))

;;; determines the position (index) of atom a in the list b
(define (list-position a b)
     (let loop ((i 1) (x b))
       (if (pair? x)
         (if (eq? a (car x)) i
             (loop (+ i 1)(cdr x)))
         #f)))

;; makes a vector and initializes its elements using the function f
(define (make&init-vector size f)
  (let
    ((w (make-vector size)))
    (do
      ((i 0 (+ 1 i)))
      ((= i size) w)
      (vector-set! w i (f i)))))

;curries 
(define (curry f a) (lambda l (apply f (cons a l))))
(define (curryn f l) (lambda l2 (apply f (append l l2))))


;;-----------------------------
;; Loads a list of files. The full path for each file is obtained by
;; concatenating path, with each element of namelist, and giving it the 
;; termination termina. NameList is a list of symbols!!.

(define (load* path namelist termina)
  (cond
   ((null? namelist) #t)
   ((pair? namelist)
    (let
      ((nom (symbol->string (car namelist))))
      (display (string-append "***Loading:" nom))
      (load (string-append path nom termina)) 
      (newline)
      (load* path (cdr namelist) termina)))
   (else 
    (let
      ((nom (symbol->string namelist)))
      (display (string-append "***Loading:" nom))
      (load (string-append path nom termina)) 
      (newline)))))

;; the path to Macgambit
(define gambit-path (string-append mac#home-folder ":"))

; Require and Provide
;--------------------
;Uses the list of paths ((name path)...) and the list of files
; of the form ((symbol (path filename))...)

(define (path x)
  (let
    ((z (assoc x *index*)))
    (if z (cadr (assoc (caadr z) *directories*)) #f)))

(define (file x)
  (let
    ((z (assoc x *index*)))
    (if z (cadr(cadr z)) #f))) 

(define *terminator* ".scm")

(define (reset-modules)
  (set! *list-of-modules* '()))

(define (provided module)
  (set! *list-of-modules* (cons module *list-of-modules*)))

(define (require module)
  (if (member module *list-of-modules*)
    #t
    (let
      ((x (path module))
       (y (file module)))
       (if (and x y) 
         (begin
           (load* x y *terminator*)
           (provided module))
         #f))))

 (reset-modules)

;----------------------------------------------------------------------


; Input-Output
;-------------

; Allow multiple argument form of display with and without final newline

(define (DisplayLine . someArgs)
    (for-each (lambda (arg)(display arg) (display " ")) someArgs)
    (newline))

(define (DisplayList . someArgs)
    (for-each (lambda (arg)  (display arg) (display " ")) someArgs))

(define (Load-And-Make-List in-port)
  (define(list-of-things)
      (let
        ((x (read in-port)))
        (if (eof-object? x) '() (cons (eval x) (list-of-things)))))
  (list-of-things))

;pathname relative to home folder
(define (pathname xx)
  (string-append mac#home-folder xx))

; Random Numbers
; --------------
; This function is suitable only for noncritical uses. Use Marsaglia if 
; more power is needed.

(define (make-random)
  (##declare (fixnum a x ##modulo))
  (let 
      ((a 37997)(x 314159293))
      (lambda() (set! x (##modulo (* x a) 2147483648)) 
                (/ (exact->inexact x) 2147483648 ))))

(set! *random* (make-random))

(define (myRandom n) (truncate (* n (*random*))))
; Returns a pseudo-random nonnegative integer less than a
; positive integer n.

(define Plus-Infinity 1.0s30)
(define Minus-Infinity -1.0s30)
(define (filter-number x epsi) (if (<= (abs x) epsi) 0 x))
;;----------------------------------
;;Numerical Formats
;;----------------------------------

(define (format numero ndecs)
  (let
    ((num (expt 10 ndecs)))
    (/ (truncate (* numero num)) num)))

(define (formatf x m n)
  (let ((numero (number->string (format x n))))
    (if (< m (string-length numero))
      (make-string m #\*)
      (string-append (make-string (- m (string-length numero)) #\ ) numero))))

(define (format-all m n . lista)
  (map (lambda(z)(formatf z m n)) lista))

; string formatting
(define (format-str x n)
  (substring (string-append x (make-string n #\ )) 0 n))
;;---------------------------------
;; Macro expander
;;---------------------------------
; a macro to help in debuging macros. Expands a macro and returns the resulting 
; expansion. Uses the same macro-expander del sistema

(defmacro (macro-expand sexpr)
      `(apply (cddr (assq ',(car sexpr) ##global-env-macros))
              ',(cdr sexpr)))
