;;; This is the high speed version of SimmerEvents. This is achieved by
;;; using functions instead of methods, avoiding the dispatching process,
;;; inlining references, using fixnum operations whenever possible, disabling 
;;; array boundary checking, using global variables to comunicate values
;;; instead of passing them as arguments, and last, but certainly not least, by 
;;; addressing objects using their internal representation as vectors. Some of 
;;; those practices, specially the last one, are VERY dangerous!!!
;;;
;;; BEWARE OF CHANGES IN MEROONs INTERNAL OBJECT REPRESENTATION, for instance
;;; when the merron.tag field was introduced find-position had to be changed.
;;; Macros to remove boundary checking in vectors
;;;

(##define-macro (+-> x i z)`(##vector-set! ,x ,i ,z))
(##define-macro (<-+ x i)`(##vector-ref ,x ,i))


;;; To decrease the danger in addressing fields of objects from their internal 
;;; representation we use macros. The macros compute the position of each field 
;;; in the object by finding its position among the object fields. To this end 
;;; we use the following function:
;;;
;;; Changed to take care of the merron.tag field!!!

(define (find-position datafield-name dataclass)
  (+ 1
  (list-position datafield-name (map field-name (class-fields dataclass)))))

(##define-macro ($object-class? x clase) 
          `(= (<-+ ,x 0)
              (<-+ ,clase ,(find-position 'number class-class))))
; -------------------------- EVENTS------------------------------------------

(define-class Event Object (Time Message Data Priority))

;;; This is the main generic that performs an event.

(define-generic (make-happen (anEv Event)))


;;; The following variables increase the efficiency of the event dispatching 
;;; process. They will be filled in $immerControl. Here we place their 
;;; definitions for loading and compilation purposes.
;;;

(define *LAP* #f)
(define *EVENT-DATA* #f)
(define *CLOCK* 0)
(define *READY-LIST #f)

(##define-macro ($set-event-dispatcher-clock! x) `(set! *CLOCK* ,x))
(##define-macro ($event-dispatcher-clock)  `*CLOCK*)


;;;----------------------------------------------------------------------------
;;; 
;;; Macros for inlining events
;;;
;;; The following stuff makes event processing more efficient. All comparisons 
;;; and all time arithmetic is in integers. Since 
;;; internal time is multiplied by 1000 and the maximum inexact number is 
;;; 2^31 = 2147483648 then this gives a maximum time of 2147483.648, more
;;; than enough!!!
;;;----------------------------------------------------------------------------

(##define-macro ($event-time elem)
                `(<-+ ,elem ,(find-position 'time     event-class)))
(##define-macro ($event-priority elem)
           `(<-+ ,elem ,(find-position 'priority event-class)))
(##define-macro ($event-message elem)
           `(<-+ ,elem ,(find-position 'message event-class)))
(##define-macro ($event-data elem)
           `(<-+ ,elem ,(find-position 'data event-class)))
(##define-macro ($set-event-time! elem x)
                `(+-> ,elem ,(find-position 'time event-class) ,x))
(##define-macro ($set-event-priority! elem x)
           `(+-> ,elem ,(find-position 'priority event-class) ,x))
(##define-macro ($set-event-message! elem x)
           `(+-> ,elem ,(find-position 'message event-class) ,x))
(##define-macro ($set-event-data! elem x)
           `(+-> ,elem ,(find-position 'data event-class) ,x))

;;;TIME

(##define-macro ($time-round time) 
                `(##inexact->exact (##round (* ,time 1000.))))
(##define-macro ($Current-Time) `(/ ($event-dispatcher-clock) 1000.))
(##define-macro ($user-event-time aEv) `(/ ($event-time ,aEv) 1000.))
(##define-macro ($Time-from-now t)`(+ ($current-time) ,t))
(##define-macro ($Next-event-Time) `($user-event-time (peek *LAP*)))
(define (current-time) ($current-time))
(define (time-round time)($time-round time))
(define (time-from-now t)($time-from-now t))
(define (user-event-time aEv)($user-event-time aEv))
(define (next-event-time) ($next-event-time))
;;;
;;; An event with priority <= to *ignore-value* is ignored!!!!. This is a
;;; way of removing events from the LAP. 

(define *ignore-value* -1)

;;; define a method to order events. They are ordered first by time and then 
;;; by priority.

(define-method (object<? (aE1 event)(aE2 event))
   (let ((x ($event-time aE1))
         (y ($event-time aE2)))
   (if (##fixnum.= x y)
     (< ($event-priority aE1)($event-priority aE2))
     (##fixnum.< x y))))


;;; a Method to print the event information.Does not pay to accelerate.

(define-handy-method (display-object (anEv Event) . stream)
    (DisplayList  " Ev:" (format ($user-event-time anEV)  3) " Pr:" priority 
                  " Mess:" message " Data:")
    (for-each show data)
    (newline))
;-------------------------RECURRENT EVENTS----------------------------------


(define-class Recurrent-event Event
        ((= recurrence-time :initializer (lambda()0.))))

(define-method (initialize! (aRE recurrent-event))
   (set-recurrent-event-recurrence-time! aRE 
        (exact->inexact (recurrent-event-recurrence-time aRE))))

(define-generic (start-object-process (anObj object) t0 . args))

(##define-macro ($recurrent-event-recurrence-time elem)
    `(<-+ ,elem ,(find-position 'recurrence-time recurrent-event-class)))


;;;------------------------RANDOM RECURRENT EVENTS -------------------------
;;; This allows for variable intervals between events given by a function in
;;; the slot recurrence-time. Notice the extra set of parenthesis around the
;;; recurrence-time slot!
;;; The event expects a function in this slot, that will be evaluated to 
;;; determine the interval to the next event.

(define-class Random-Recurrent-event Recurrent-Event ())

(##define-macro ($random-recurrent-event-recurrence-time elem)
    `(<-+ ,elem ,(find-position 'recurrence-time random-recurrent-event-class)))

;;;--------------------support functions for event management ----------------
;;; The following functions help in creating and managing events. Every function
;;; returns the event, which is useful to locate events if you keep track of it.

;;; internal function that enters events. Event time is rounded to prevent silly
;;; effects. It returns the event.

(define (event time message arg pr)
  (let ((ev (make-event ($time-round time) message arg pr)))
    (if ($put-event! ev) ev
      (error "No room in Pending Events List when loading events"))))

;; creates absolute events.
(define (create-abs-event time message . arg)
  (event time message arg 0))

; creates events relative to current time.
(define (create-event time message . arg)
  (event ($time-from-now time) message arg 0))


;; event dispatcher scheduling relative to current-time
(define (schedule-resume aP time . args)
      (event ($time-from-now time) smc:resume (##cons aP args) 0))

;; absolute dispatcher scheduling. Use with a lot of care!!!
(define (schedule-resume-abs aP time . args)
   (event time smc:resume (##cons aP args) 0))

; creates events from a list of tuples of the form (time object message . data)
(define (load-events event-list)
  (for-each (lambda(x)(apply create-event x)) event-list))

; creates events allowing the assignment of priorities
(define (create-event-with-priority time message pr . args)
  (event ($time-from-now time) message args pr))

; Creates recurrent events
; of the form <initial delay> <Priority> <time-between-events> <message> <data>
; all events derived from a recurrent event have the same number

(define (create-recurrent-event time Priority Rec-time message . arg)
  (if (exact? priority)
    (let ((ev (make-recurrent-event 
             ($time-round ($time-from-now time)) message arg  
             priority rec-time)))
    (if ($put-event! ev) ev
      (error "No room in Pending Events List when loading events")))
    (error "Invalid Priority " priority)))

;=============================================================================
;;;  IMPATIENCE
;=============================================================================
;;; The following procedures form the skeleton of the impatience mechanism. 
;;; It is very elementary and not foolproof. 
;;; You can always interrupt a wait *not in the event list*  by scheduling an
;;; impatience event that starts some alternative path. However you have to
;;; access to the LAP to find the initial scheduled event that was supossed to
;;; terminate the wait in a normal way. This can be done by
;;; 1.- Keeping track of the event, then using schedule-resume to generate
;;;     the impatience event and put it into the LAP. Then start the wait.
;;; 2.- Once the reawakening of the process happens, you should test
;;;     for the event to see if it is a normal or impatience exit. You
;;;     can recover the last-event executed by using the get-last-event function
;;;     An impatience exit will be signalled by the impatience event. Then 
;;;     perform the impatience code.
;;; 3.- If normal then you should remove the impatience event from the LAP. You
;;;     can do it by using ignore-event.
;;; 
;;; This can be used for all waits for conditions, but not for waits for a 
;;; given duration. This is because for an scheduled wait that has an associated
;;; impatience event, any one happening requires removing the other. This
;;; in turn requires knowing the normal event, and 
;;; removing it if impatience occurs. This can be done by the user, but you 
;;; have to bypass wait-duration because it does not return the event (it can 
;;; not because wait-duration becomes suspended, and does not return)           ;;; Perhaps a new macro culd be defined along the lines in the exemple 11.
;;; It is left as a future extension, if needed...

;;; The function ignore-event marks the priority of a supplied event to 
;;; *ignore-value*. When *overlord* finds a marked  priority ignores the event. 
;;; Thus ignore-event efectively suppreses an event.

(define (ignore-event ev)(set-event-priority! ev *ignore-value*))

(define (get-last-event)(event-dispatcher-last-event *overlord*))




;;;****************************************************************************
;;;
;;;            Heap handling to accelerate event processing
;;;
;;;****************************************************************************

;;; and this is the macro used internally
(##define-macro ($object<?-event x y)
   `(if (##fixnum.< ($event-time ,x)($event-time ,y)) #t
        (if (##fixnum.= ($event-time ,x)($event-time ,y))
          (##fixnum.< ($event-priority ,x)($event-priority ,y))
          #f )))
  

(##define-macro (nEVS)      `(<-+ *LAP*   ,(find-position 'n    heap-class)))
(##define-macro ($set-nEVS! x)`(+-> *LAP*   ,(find-position 'n  heap-class) ,x))
(##define-macro (LAP-size)  `(<-+ *LAP*   ,(find-position 'size heap-class)))

          
;;;
;;;---------------Event Heap--------------------------------------------------
;;; This is a specialization of put! and get! to give the maximum speed in
;;; handling event heaps. 
;;;
;;; This is the equivalent of put! heavily specialized to the LAP
;;;
(define ($put-event! args)
 (##declare (fixnum))
 (let ((n (nEvs))) 
  (define (insert j)
    (if (##fixnum.= j 1) #t
        (let ((x (<-+ *EVENT-DATA* j))
              (i  (quotient j 2)))
          (if ($object<?-event  x (<-+ *EVENT-DATA* i))
            (begin
             (+-> *EVENT-DATA* j (<-+ *EVENT-DATA* i))
             (+-> *EVENT-DATA* i x)
             (insert i))))))
  
    (if (##fixnum.= n (LAP-size)) #f
      (let (( n1 (##fixnum.+ 1 n)))
        (+-> *EVENT-DATA* n1 args)
        (insert n1)
        ($set-nEVS! n1)
        #t))))

;;; This is the equivalent of get! for the LAP. Watch out for recurrent events.
;;; To increase their efficiency, we do not remove a recurrent event, and we
;;; just replace it and sift as you can see in the next function
;;;
(define ($get-event!)
  (##declare (fixnum))
  (let ((n (nEvs)))
    (let ((x (<-+ *EVENT-DATA* n)))
      (define (combina i n)
        (let ((j (##fixnum.* 2 i))
              (m n))
          (if (##fixnum.> j m)  
            i
            (begin
             (if (and (##fixnum.< j m) 
                      ($object<?-event (<-+ *EVENT-DATA* (##fixnum.+ 1 j)) 
                               (<-+ *EVENT-DATA* j)))
               (set! j (##fixnum.+ 1 j)))
             (if ($object<?-event  (<-+ *EVENT-DATA* j) x)
               (begin 
                (+-> *EVENT-DATA* i (<-+ *EVENT-DATA* j)) 
                (combina j n))
               i)))))
      
      (if (##fixnum.= n 0) #f
          (let ((n1 (##fixnum.- n 1)))
            (+-> *EVENT-DATA* (combina 1 n1) x)
            ($set-nEVS! n1))))))


;;; This is for events that were not removed from the LAP. Recurrent events
;;; are simple changed. Therefore it is simpler to just sift the changed 
;;; event to its right place instead of removing and then adding it.

(define ($replace-event! x)
 (##declare (fixnum))
  (define (sift)
    (let ((r (nevs)))
      (let loop ((i 1) (j 2))
        (if (##fixnum.> j r) i
            (begin 
             (if (and (##fixnum.< j r) 
                      ($object<?-event (<-+ *EVENT-DATA* (+ j 1))
                               (<-+ *EVENT-DATA* j))) 
               (set! j (##fixnum.+ j 1)))
             (if ($object<?-event x (<-+ *EVENT-DATA* j)) i
                 (begin
                  (+-> *EVENT-DATA* i (<-+ *EVENT-DATA* j))
                  (loop j (##fixnum.* 2 j)))))))))
  (+-> *EVENT-DATA* (sift) x))

