;(##include "meroon.gam")
;(##include "macros.scm")
;(##include "careless.scm")
;(##include "macinst.scm")
;(##include "CompilationStuff.scm")

(require 'vector-arithmetic)
(require 'simulation)
(require 'graph-vars)
(require 'monitors)

;A virtual class that defines the basic plant structure.
;
(define-class dynamic-object named-object
     ((= controller :initializer (lambda()#f))
      (= receiver   :initializer (lambda()#F))) :virtual)

(define-method (initialize! (aDO dynamic-object))
   (set-dynamic-object-stoped! aDO 
        (instantiate condition 
                     :parking-place #t 
                     :name (qualify-name aDO ".condition")))
   aDO)


(define-generic (dynamic-initialize (aDF dynamic-object) . args))
(define-generic (simulate-step (aDF dynamic-object)))
(define-generic (input-map (aDF dynamic-object) . args))
(define-generic (output-map (aDF dynamic-object) . args))


(define-method (stop-object (aDF dynamic-object))
   (reset (dynamic-object-stoped! aDF)))

(define-method (start-object (aDF dynamic-object))
   (signal (dynamic-object-stoped aDF)))

;
; The class dynamic-object is a virtual class and its here to pull 
; functionality. It should be subclased but not instantiated.
;-------------------------------------------------------------------------
(define-class continuous-object dynamic-object
  ((= active :initializer (lambda()#t))
   (= ni :initializer (lambda()0))        ;dimension de los inputs
   (= no :initializer (lambda()0))        ;dimension de los outputs
   (= outputs :maybe-uninitialized)       ;vector de outputs
   (= inputs :maybe-uninitialized)        ;vector de inputs
   (= t :initializer (lambda()0))         ;tiempo de integracion
   (= incrt :initializer (lambda()0.1))   ;incremento entre inputs
   (= inct :initializer (lambda()0.01))): virtual) ;step-size

;crea los vectores de estado y de input y output indicados por las dimensiones
(define-handy-method (initialize! (aDO continuous-object))
  (call-next-method)
  (set! incrt (exact->inexact incrt)) ; make sure is a floatnum.
  (if (not (zero? ni))(set! inputs (make-vector ni 0.)))
  (if (not (zero? no))(set! outputs (make-vector no 0.)))
   aDO)
 


;;;-------------------DIFFERENTIAL OBJECTS-------------------------------------
; A differential object is a system of differential equations. La estructura de
; un objeto diferencial es  dX/dt = v(X,t,y)
; X es el vector de estado, cuyo nombre es state y dimension ns
; v es el metodo velocity-map que recibe como oargumentos X, t y el objeto
;         completo, que incluye los inputs.
; y es el vector de inputs de dimension ni.

; La relacion de un sistema con el exterior se realiza exclusivamente a traves
; de los vectores de inputs y outputs. Ningun metodo debe intentar alterar el
; vector de estado de las ecuaciones, que es totalmente interno y reservado.
; Idealmente, toda comunicacion al exterior, de salida, debera producirse a
; traves del vector de outputs. 

; input-map pone en el vector de inputs los valores necesarios para el calculo
;           de los segundos miembros a traves de velocity-map
; output-map pone en el vector de outputs los datos que se desean transmitir al
;           resto del mundo. Nadie mas debe tener acceso al vector de estado.
;
; Todo esto es pura convencion. Caben muchas interpretaciones por parte 
; del usuario. En realidad estos dos mapping son la tipica construccion ante y 
; post. Un metodo before se llama antes de un proceso y un after despues.
; Si se mira bien input-map y output-map son un metodo before y un after,
; que se pueden llamar antes y despues de cada paso de integracion y uno puede 
; hacer lo que quiera con ellos.
;
; La especificacion de un sistema de ecuaciones diferenciales, por lo tanto
; debe constar de los tres maps velocity-map, input-map y output-map. 

; La integracin se desarrolla a tres niveles, por medio de los metodos
; integration-method, integrate-step e integra.
; El metodo integration-method hace la integracin sobre un intervalo inct,
;    intervalo elemental de integracin. Utiliza un metodo de Runge-Kutta de
;    cuarto orden, con step-size inct.
;    El metodo llama repetidas veces al metodo velocity-map para calcular el 
;    segundo miembro del sistema de ecuaciones. En principio velocity map puede
;    hacer cualqueir cosa ya que tiene todo el sistema a su disposicin.
;    Sin embargo se espera que velocity-map use slo los valores del estado,
;    del tiempo, y de los inputs externos contenidos en el vector de inputs.
;    Esta convencin, no llamando al metodo input-map, corresponde a la 
;    estructura para la que se ha pensado el esquema de integracin.
; El metodo integrate-step, integra las ecuaciones diferenciales hasta el 
;    tiempo indicado, a partir del valor del tiempo actual. Durante esta
;    integracin se llama al metodo anterior. Si el metodo anterior sigue la
;    convencion de no llamar a  , se integra sin cambiar el valor de
;    los inputs, manteniendolos contantes durante el intervalo en cuestion.
; El metodo dynamic-initialize inicializa los datos del objeto en el momento de
;    empezar su simulacion. Es la forma tipica de tratar todas las cosas 
;    que tienen asociado un tiempo, como por ejmeplo los valores iniciales 
;    de las graficas. 

; Este sistema, se hace en prevision de su aplicacion a sistemas con
; inputs muy elaborados y de calculo complicado, pero que varian lentamente 
; en el tiempo con respecto al intervalo elemental de integracin.
;
; El metodo delta-map permite el cambio instantneo del estado del sistema,
; De hecho delta map puede hacer lo que se quiera. Se define una generic 
; delta-state-map para facilitar el proceso, cuando el efecto sobre el estado 
; es sumar a este una cierta funcion del input. delta-state-map es entonces
; esta funcion. Si el comportamiento es diferente entonces hay que redefinir
; el metodo delta-map. El argumento i del delta map especifica el input que
; recibe la delta function.

;----------------------------------------------------------------------------- 

(define-class differential-object continuous-object
   ((= ns :initializer(lambda()1))         ;dimension del estado
    (= state :maybe-uninitialized)))         ;vector de estado)

(define-generic (velocity-map (aDF differential-object) x t))
(define-generic (delta-map (aDF differential-object) x i))
(define-generic (delta-state-map (aDF differential-object) x i)  x)
(define-generic (integration-Method(aDF differential-object) x t0))


(define-method (initialize! (aDO differential-object))
   (call-next-method)
   (set-differential-object-state! aDO 
           (make-vector (differential-object-ns aDO) 0.))
    aDO)

; realiza la integracin elemental, un paso de Runge-Kutta de orden 4.
(define-handy-method 
 (Integration-method (aDO differential-object) x t0)    
  (let*
    ((inct2 (/ inct 2.0))
     (k1 (aTimesK inct (velocity-map aDO x t0)))
     (k2 (aTimesK inct (velocity-map aDO (aplus x k1) (+ t0 inct2))))
     (k3 (aTimesK inct (velocity-map aDO (aplus x k2) (+ t0 inct2))))
     (k4 (aTimesK inct (velocity-map aDO (aplus x (atimesK 2 k3)) 
                                         (+ t0 inct)))))
    (atimesK 0.166666666666 (aplus k1 k2 k2 k3 k3 k4))))

;el metodo de euler!!!
(define-handy-method
  (Integration-method (aDO differential-object) x t0)    
  (aTimesK inct (velocity-map aDO x t0)))

(define-method (velocity-map (aDO differential-object) x t0)
   (Error "Method velocity-map" " To be defined in subclass"))
;
;----------------------------Los METODOS EXTERNOS--------------------------

(define-handy-method (simulate-step (aDO differential-object))
  (set! state (aplus state (integration-method aDO state t)))
  (set! t (+ t inct))
  #t)

; el metodo delta map se responsabiliza de tratar delta-inputs, que le vienen
; directamente como argumentos. La idea basica es que le lleguen directamente
; desde un acontecimiento programado por alguien. Requiere un mapping entre 
; el input y el estado, delta-state-map para convertir la seal de entrada en 
; un cambio instantaneo del estado. El default es la identidad.
; el metodo default de delta-map lo que hace es recibir un vector(o un numero) 
; en x0 y sumarlo al componente i-esimo del estado del sistema.

(define-method (delta-map (aDF differential-object) x0 i)
 (let ((x (delta-state-map aDF x0 i)))
   (set-differential-object-state! aDF 
                                   (aplus (differential-object-state aDF) x))))

(define-handy-method (dynamic-initialize (aDO differential-object) . args)
    (set-differential-object-t! aDO (current-time))) 

;;--------------------A PROTOTYPE OF EXECUTIVE --------------------------
;; For any object, first create events to perform dynamic initialization.
;; Then create recurrent events to do the processing. 
;; Remember that the arguments of create recurrent event are
;;    - the time for the first events of the class.
;;    - the priority
;;    - the interval of time between succesive events of the class
;;    - the method to be called (or procedure to be executed)
;;    - the object (or arguments to the procedure)

(define-handy-method (start-object-process (aDO differential-object) t0 . args)
    (create-event t0 dynamic-initialize aDO)
    (create-recurrent-event  t0 100 inct  simulate-step aDO))
