;;; Meroon utilities
;;;ooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;; Here we provide the following functions, that supplement some of the basic
;;; capabilities in Meroon. I have reported all of them to Christican Queinnec
;;; but these ones failed to make it into the standard, because either they
;;; are unsafe, too specific for simmer, or difficult to make portable.

;;; Macgambit
;;;----------
;;; edit-files         opens an editor window on each file in the list

;;; Information
;;;---------------
;;; show-method        shows the source code for a method
;;; show-class-name    shows the class name from the class number
;;; show-fields        shaws the name of all fields of a class
;;; get-name           gets the name of an object

;;; Conversion
;;;------------
;;; typecast           forces the class of an object
;;; polyf->vector      converts a poly-field into a vector
;;; vector->polif      puts a vector into a poly-field

;;; Predicates
;;;-----------
;;; field-uninitialized?   #t if a field is uninitialized
;;; is-there-a-field-in?   #t if the object has a field of the given name
;;; is-generic?            #t if the name corresponds to a generic function

;;; Other
;;;------
;;; qualify-name     adds qualifiers to an object name
;;; shallow-unveil   first level unveil



;;;--------------------------------------------------------------------------- 
;;; auxiliary function for Gambit. Opens a window to edit a file
;;;--------------------------------------------------------------------------- 
 
(define (edit-files path lista )
  (if (pair? lista)
    (let ((lista0 (if (symbol? (car lista)) (map symbol->string lista) lista))
          (dir0 (if (symbol? path) (symbol->string path) path)))
      (map (lambda(x)
             (let ((x0 (string-append path x ".scm")))
               (mac_#edit x0 0 0))) 
           lista0))))

;;;--------------------------------------------------------------------------- 
;;; Some auxiliary functions for meroon
;;;--------------------------------------------------------------------------- 
;;; Gets and shows the source for the given method

(define (show-method gener class)
     (pp (find-method1 (vector-ref (->generic gener) 5)
                 (class-number (->class class)))))

;;; Forces the class type to the desired one ***without*** changing the class
;;; allocation. Useful to call methods several classes upwards!.
;;; Careful this is dangerous stuff. Remember to reset it back if you want to do
;;; other things with the object

(define (typecast x clase-destino)
  (-> x 0 (class-number (->class clase-destino)))
  x)

;;; Turns a polyfield into a vector and returns it.

(define (polyf->vector objeto name)
  (let* ((n (field-length objeto name))
         (x (make-vector n)))
    (for i 0 (- n 1) (-> x i (field-value objeto name i)))
    x))

;;; sort of a converse. puts the elements of the vector unvector into the 
;;; elements of the polyfield. Returns the object.

(define (vector->polif objeto name unvector)
  (let* ((n (field-length objeto name)))
    (for i 0 (- n 1) (-> x i (set-field-value! objeto name i (<- unvector i))))
    objeto))

;;; tests if the field named "name" in the object is uninitialized
(define (field-uninitialized? obj name)
  (uninitialized?
         (instance-ref obj 
              (compute-offset obj 
                    (retrieve-named-field (object->class obj) name)))))

;;; shows the name of the class given its number
(define (show-class-name numero)(class-name (->class numero)))

;;; find a field with the given name or return #f

(define (is-there-a-field-in? y x)
    (memq x (show-fields (class-name (object->class y)))))

;;; Utility to get the name,as a string. of any object that has such a field  
;;; Works even if unnamed.

(define (get-name anObj)
  (if (is-there-a-field-in? anObj 'name)
    (if (field-uninitialized? anObj 'name)
      (string-append "UnNamed-" 
                     (symbol->string (class-name (object->class anObj))))
      (if (symbol?(field-value anObj 'name))
        (symbol->string (field-value anObj 'name))
        (field-value anObj 'name)))
    (error "Name not defined in class"(class-name (object->class anObj)))))


;; creates a composite name for an object by adding extras to it. Extras can be
;; any simple type. This is the difference between this and symbol-concatenate

(define-generic (qualify-name (x object) . extra)
     (let* ((trans (lambda(y)
                    (string-append "-"
                                   (cond ((symbol? y)(symbol->string y))
                                         ((number? y)(number->string y))
                                         ((char? y)(char->string y))
                                         ((string? y) y)
                                         (else ""))))))
            (apply string-append (get-name x) (map trans extra))))

;;; shows the names of the fields in class x (x should be a name)
(define (show-fields x)(map field-name (class-fields (->class x))))

;;; tests if a function is generic
(define (is-generic? x)(generic? (->generic x)))

;;;--------------------------------------------------------------------------- 
;;;  Shallow unveiling
;;; half borrowed and adapted from Queinnec
;;;--------------------------------------------------------------------------- 
;;; a kind of unveil, but first level only. The problem with unveil in big
;;; objects is too much detail. Here we just show the first level of the object
;;; and show (non recursively) the values of its fields

;;; This is the main function:

(define (shallow-unveil o . stream)
  (let ((stream (if (pair? stream) (car stream) (current-output-port))))
    (shallow-show-unveiled o stream)
    #t ) )

  
;;; Pretty print O

(define (shallow-show-unveiled o stream) 
  (define (shallow-show-object-unveiled o stream)
    (let ((class (object->class o)))
      ;;(format stream "(a ~A <------------- [Id: ~A]"
                ;;        (Class-name class) index )
      (display "(a " stream)
      (display (format-str (get-name class) 20) stream)
      (for-each (lambda (field) 
                  (shallow-show-unveiled-field-content
                    o stream field ) )
                (Class-fields class) )
      ;;(format stream " end ~A)" (Class-name class))
      (newline stream)
      (display " end " stream)
      (display (Class-name class) stream)
      (display ")" stream) ) )
  
  (if (Object? o) 
    (shallow-show-object-unveiled o stream)
    (show o stream)))


;;; Display the content of a field according to the type of the field.
;;; This is a generic function you can extend on new types of fields
;;; if any.

(define-generic (shallow-show-unveiled-field-content 
                  o stream (field)))


;;; Shallow Show a Mono-Field value

(define (show-value-or-object xxx stream)
  (define (saca-lista lista)
    (display "#(" stream ) 
    (for-each (lambda(x)
                (show-value-or-object x stream)
                (display #\  stream)) 
              lista)
    (display #\) ))
  
  (cond ((vector? xxx)
         (if (object? xxx) 
           ;it seems to be an object
           (show xxx stream)
           (saca-lista (vector->list xxx))))                      
        ((list? xxx) (saca-lista xxx))
        (else (display xxx stream))))
        
(define-method (shallow-show-unveiled-field-content 
                 o stream (field Mono-Field) )
  (newline stream)
  (display "  -From: " stream)
  (display (format-str (get-name (->class (field-class-number field))) 20)
           stream)
  (display (Field-name field) stream)
  (display ": " stream) 
  (if (field-defined? o field)
   (show-value-or-object (field-value o field) stream)
   (display "#<Uninitialized>" stream) ) )

;;; Shallow Show a Poly-Field value

(define-method (shallow-show-unveiled-field-content 
                 o stream (field Poly-Field) )
  (newline stream)
  (display "  -From: " stream)
  (display (format-str (get-str-name (->class (field-class-number field))) 20)
           stream)
  (let loop ((n (field-length o field))
             (i 0) )
    (when (< i n)
      (display (Field-name field) stream)
      (display "[" stream)
      (display i stream)
      (display "]: " stream)
      (if (field-defined? o field i)
        (show-value-or-object (field-value o field) i stream)
        (display "#<Uninitialized>" stream) )
      (loop n (+ 1 i)) ) ) )

