
(require 'named-objects)
(require 'stat-vars)
;
;--------------------------------------------------------------------------
;;; General Class COLLECTION
;; This is an abstract class not to be instantiated.
;; Most of the methods in this class require a generating function
;; that produces one element of the collection every time that is called
;; until the collection is exhausted. Then it returns #f. The order of the
;; returned elements is the natural order for the class, if the class is ordered
;; Otherwise it is arbitrary. The function should return each element exactly 
;; once
;; The generating function is internally produced by calling the method
;; make-get-elt on an object of the class. The returned function should have 
;; no arguments.
;; Each subclass should provide its make-get-collection method, and its get!
;; and put! methods.
;; put! puts an element in the collection
;; get! removes and returns the next element, and should be defined only when
;;      "next" has a meaning. 

(define-class collection named-object () :virtual)

;; get! and put! are defined in subclasses

(define-generic (put! (as Collection) arg))
(define-generic (is-in? (as Collection) arg))
(define-generic (empty? (as Collection)))
(define-generic (any-elt? (as Collection) predicate))
(define-generic (every-elt? (as Collection) predicate))
(define-generic (for-each-elt (aS Collection) procedure))
(define-generic (map-elt (aS Collection) predicate))
(define-generic (fill (as Collection) lista))
(define-generic (make-get-elt (as Collection)))
(define-generic (expand-size (as Collection) . amm))
(define-generic (peek (aOC collection)))
;;;
;;; generic with default behavior for Scheme objects
;;;
(define-generic (get! (as Collection)) aS)

;;tests to see if an element belongs to the collection
(define-method (is-in? (as Collection) arg)
    (any-elt? as (lambda()(equal? arg x))))

;;returns the element if some element in the collection returns #t for 
; the given predicate
;;#f otherwise. Does short circuit evaluation.
(define-method (any-elt? (as collection) predicate)
  (let
    ((els (make-get-elt as)))
    (let loop
      ((x (els)))
      (cond 
       ((not x) #f)
       ((predicate x) x)
       (else (loop (els)))))))
 
;;returns #t if all elements in the collection return #t for the predicate
;;supplied. #f otherwise.                                  
(define-method (every-elt? (as collection) predicate)
    (not (any-elt? as (lambda(x)(not(predicate x))))))

;; applies the procedure to each element of the collection.
(define-method (for-each-elt (as collection) procedure)
   (let
     ((els (make-get-elt as)))
     (let loop
       ((x (els)))
       (cond 
        (x (procedure x)(loop (els)))
        (else #t)))))

;;same, in the order given by make-get-elt and returns a list of results.
(define-method (map-elt (as collection) procedure)
    (let
      ((els (make-get-elt as)))
      (define (inner-map)
        (let
          ((y (els)))
          (cond 
           (y (cons (procedure y)(inner-map)))
           (else '()))))
      (inner-map)))
                 

;; insert into the collection all members of the list lista.
(define-handy-method (fill (q Collection) lista) 
   (for-each (lambda(x)(put! q x)) lista)
   #t)

;==========================================================================
;;------------------Some special Collections-------------------------------
;
;; keyed-collections store objects that answer to the message "key". 
;; The generic key defines keys for all default scheme objects.

(define-class keyed-collection collection () :virtual)

;;the special (primitive) generics to handle records with keys
;-----------------------------------------------------------
(define-generic (remove-elt (aKC keyed-collection) key)
    (error "remove-elt has to be defined in a subclass" key))

; default assumes numeric keys and performs a sequential search!!!.

(define-generic (fread (aKC keyed-collection) key))
(define-method (fread (aKC keyed-collection) key)
    (let ((els (make-get-elt aKC)))
      (let loop ((x (els)))
        (if x 
          (if (= (object-key x) key) x (loop (els)))
          #f))))
      
;--------------a simple collection illustrating class------------------------

(define-class keyed-list keyed-collection 
      ((= data :initializer (lambda()'()))))

(define-method (put! (aKL keyed-list) arg)
   (set-keyed-list-data! aKL (cons arg (keyed-list-data aKL))))

(define-method (fread (aKL keyed-list) key)
   (define (aux-read lis)
     (if (pair? lis) 
       (if (equal? (object-key (car lis)) key) (car lis)
                                               (aux-read (cdr lis)))
       #f))
   (aux-read (keyed-list-data aKL)))

(define-method (remove-elt (aKC keyed-list) key)
     (let ((list (keyed-list-data aKC)))
       (cond ((null? list) #f)
             ((equal? (object-key (car list)) key) 
              (set-keyed-list-data! akc (cdr list)))
             (else 
              (let loop ((first list)
                         (second (cdr list)))
                (if (equal? (object-key (car second)) key) 
                  (begin
                   (set-cdr! first (cdr second)))
                  (loop second (cdr second))))))))

(define-handy-method (make-get-elt (aSKL keyed-list))
  (let
    ((lista data))
    (lambda()(if (pair? lista)
               (let
                 ((w (car lista)))
                 (set! lista (cdr lista))
                 w)
               #f))))


(define-class stat-keyed-list keyed-list
   ((= number :maybe-uninitialized)))

(define-method (initialize! (aSKL stat-keyed-list))
   (set-stat-keyed-list-number! aSKL
        (instantiate time-var :name (qualify-name aSKL ".estadisticas")))
   aSKL)

(define-handy-method (put! (aSKL stat-keyed-list) arg)
   (call-next-method)
   (increment number arg))

(define-handy-method (remove-elt (aSKL stat-keyed-list) key)
   (let
     ((x (call-next-method)))
     ( if x (decrement number arg))
     x))

;;;---------------------------ORDERED COLLECTIONS------------------------------
;;;

(define-class ordered-collection keyed-collection () :virtual)
