
(require 'queues)
(require 'heaps)
(require 'named-objects)
(require 'events)

;;;TIME
;;; Time is rounded to thousandth. This is to avoid trivialities produced
;;; by small round off error in floating point quantities. Time is always stored
;;; multiplied by 1000, and reduced to its propoer scale when retrieved.
                                 
(define (time-round time) (inexact->exact (round (* time 1000.))))
(define (Current-Time) (/ (Event-Dispatcher-clock *overlord*) 1000.))
(define (user-event-time aEv) (/ (event-time aEv) 1000.))
(define (Time-from-now t)(+ (current-time) t))
(define (Next-event-Time)
  (event-time (heap-peek (event-dispatcher-LAP *overlord*))))
;;;
;;;------------------------------SIMULATION CONTROL------------------------
;;; Discrete, or event oriented simulation is performed here. The core of the 
;;; system is an event dispatcher that allows scheduling and executing of
;;; events under the control of a master time clock.
;;; Both event oriented and process oriented simulation are posible, since we
;;; provide both kinds of dispatchers, a process dispatcher and an event
;;; dispatcher. This is the basic layer of the simulation control.
;;; With the file SimmerEvents It defines everything needed to control 
;;; processes and events 
; 
;;;   Events    are the basic time keeping mechanisnm. It is full generality 
;;;             an event is a procedure that will execute when its time arrives.
;;;             Events can carry with them the arguments to the calles procedure
;;;             plus many other interesting things: a unique identification
;;;             number, a priority, etc.
;;;   Processes are the basic stuff of the simulation. A process is a basic 
;;;             frame structure to store the continuation of a body
;;;             of code and allows its execution in a threaded way. The code is 
;;;             represented as a method or as a function, and when coupled with 
;;;             the process structure. Then they execute in an asynchronous way 
;;;             with the rest of the code, using continuations to switch among 
;;;             processes. Processes can be created, interrupted, suspended and 
;;;             finished, by using in the user code the primitives provided. 
;;;             Processes are normally created automatically whenever the user 
;;;             defined an async-method or an async-function. An async-method 
;;;             is a method (for any object) that has associated a process 
;;;             structure that allows for continuation processing.
;;;
;;;             Async methods and functions are, in the simplest case, exactly 
;;;             the same as normal code. The only addition is the need for the 
;;;             variable self, local to the method, that should be initialized 
;;;             to an instance of the class process.
;;;
;;;             The user can forget about this, since this is done automatically
;;;             by the macros supplied. We supply macros to create 
;;;             two kinds of async code: async-methods and async-functions. 
;;;             They are started by calling them in a normal way, and then, 
;;;             they wait in the scheduler until they are selected for execution
;;;             They do not return values at the point of invocation since they 
;;;             follow continuations, not reurning to the scheduler.
;;;
;;;             We provide waitfor methods and functions. A waitfor method is 
;;;             just another normal method, but allows the user to refer to self
;;;             from inside the body of the method, referring to the underlying 
;;;             process (the process that called the callers of the waitfor 
;;;             method). The same is applicable to waitfor functions.
;;;
;;;   Process-Dispatchers are coroutines. They have a queue, where pending
;;;             processes are stored, waiting for execution. They dispatch
;;;             (execute) the processes by taking care of the
;;;             synchronization aspects. There is a root process-dispatcher
;;;             *overlord* (actually an event-dispatcher, a subclass).
;;;             Once the dispatcher runs out of processes it dispatcher 
;;;             terminates execution and control goes to the event dispatcher 
;;;             in the same root object "*overlord*".
;;;
;;;   Event-Dispatcher. It owns a list of pending events LAP. 
;;;             It provides wake-up services to all other processes. 
;;;             Event-dispatcher is a subclass of process-Dispatcher. There is 
;;;             only a single Event-dispatcher, called *overlord*.

;;; All processes must be waiting somewhere at all times if they have to be 
;;; executed asynchronously.
;;; The running process is stored in the variable *executing*. Therefore, 
;;; it can be sent into oblivion by not puting it anywhere and invoking the 
;;; scheduler 
 
;;;
;;; To have a set of processes running under *overlord* do as follows:
;
;;; 1) Define as many objects and methods as needed by the simulation.
;;; 2) define async methods and functions  by using the macros supplied 
;;; 3) create *overlord* by doing:   (init-simulation)
;;; 4) Now, initialize all async methods and functions by calling them with
;;;    their name. 
;;; 5) Start the top level supervisor *overlord* by doing (Start-Simulation).
;;;
;;; You may want to perform additional initializations, for instance setting
;;; up the random number generator by doing (set! *random* (make-random))
;;; thus insuring the same sequence of random numbers across simulations.

;;---------CONTROLLING the SIMULATION--------------------------
;;; Checkpointing means storing the continuation of the process.
;;; The structure of control is as follows:
;;;
;;; a.- General coroutine methods:
;;;
;;;      switch : calls another coroutine saving the continuation of the caller
;;;      resume :goes to the continuation of a coroutine, 
;;;               not saving anything and returning its argument

;;; b.- Controlling processes. 
;;;     Functions that are used from the inside of a process to control its
;;;     time behavior.

;;;  a) functions not queuing the process,i.e. those methods do not allow the
;;;     dispatchers to control the flow of process activation. They are used
;;;     for threading along processes. Similar to the subroutine statement 
;;;     end.
 
;;;        abandon  :returns to the process scheduler, the process is forgotten.
;;;        interrupt: checkpoints the process and calls the scheduler 

;;;  b) functions queuing the process into a scheduler. They put the process
;;;     in a queue for control by the owner of the queue and call the process
;;;     scheduler, therefore interrupting the process.
;;;
;;;       schedule puts the process on the ready list of the scheduler.
;;;       park: puts the process in some waiting place and interrupts it
;;;       wait-duration :queue it into *overlord* and interrupt it
;;;       reschedule: puts the process in the ready-list and interrupt it.
;;;
;;;       
;;; c.-Controlling the process-dispatcher
;;; There is little need for the user to control the process dispatcher, unless
;;; he/she is involved in substantial extensions. The following methods are the
;;; most useful ones.

;;;    schedule :puts a process, supplied as argument into the queue of the PD.

;
;;; 4.-Some useful macros

;;;      (define-async-method ((name type) arguments) .  body)
;;;      (define-waitfor-method ((name type) arguments) . body))
;;;      (define-async-function ((name type) arguments) .  body)
;;;      (define-waitfor-function ((name type) arguments) . body))
;;;                       
;;;  where other is a list of pairs field value as in instantiate.                         
;
;;; 5.-Additional control/information 
;;;      (current-time)
;;;      (next-event-time)
;;;
;;; see the code in the following and in simmerevents for many useful functions
                         

;;;======================== GLOBAL DEFINITIONS=================================
;; What is in *overlord* is what receives the dispatch message, invoked to
;; get the next activity to be performed. 

(define *overlord* #f)

;;; All the controls of processes are functions and not methods, because I
;;; do not expect the switching logic to depend on the subclass. It is
;;; inmutable and the same for all switching processes.

;;; This is the process currently being executed

(define *executing* #f)


;;;======================== PROCESSES =======================================

(define-class coroutine Named-Object
     (( = continuation :initializer (lambda()#f))))

;;; a.- The basic function for coroutines
;;; The following functions expect in aP an object in a subclass of process.
;;; This is the basic function, that stop a coroutine and switches to 
;;; another.

(define (SMC:switch aP anotherC Args)
   (call/cc (lambda(cont)
             (set-coroutine-continuation! aP cont)
             (SMC:Resume anotherC args))))

;;; Processes are frames to allow the interruption of a body of code.
;;; Since they may need to keep the state they are furnished with a data slot.
;;; Only one process may executing at a time. It is stored in *executing*
;;; In general all methods need to deal with *executing*

;;; First a few constants that allow the system to know the state of a process

(define iddle 0)
(define waiting-for-event 1)
(define interrupted 2)
(define ready-to-run 3)
(define running 4)

(define-class Process Coroutine 
    ((= status :initializer (lambda()iddle))
     ( = data :initializer (lambda()#f))))

;; Processes do not have identification keys. However, since they could be
;; stored with objects in their data slot, they should inherit their key from
;; whatever is there.

(define-method (object-key (aP process))
     (object-key (process-data aP)))

(define-method (object<? (aP process)(aP1 process))
     (object<? (process-data aP)(process-data aP1)))

;;; b) Methods to handle reawakening of processes. In general they
;;;   1) Put the process somewhere (in a queue) where it can be recovered
;;;   2) call the scheduler
;;; Typical places to put a process, if you want to continue it, are the
;;; scheduler waiting-list (process-dispatcher-ready-list aK) or in the LAP

;;; This function resumes an interrupted process.

(define (SMC:Resume aP args) 
  (set! *executing* aP)
  (set-process-status! aP running)
  ((coroutine-continuation aP) args))

;;; The running process is nowhere. Therefore this forgets about it brutally.

(define (SMC:abandon) (SMC:resume *overlord* #f))

;;; Put the process in the ready-list of the dispatcher.

(define (SMC:schedule aP args)
  (set-process-status! aP ready-to-run)
  (set-process-data! aP args)
  (put! (process-dispatcher-ready-list *overlord*) aP))

;;; The main procedure for process switching. Stops the process and calls the
;;; dispatcher. 

(define (SMC:interrupt aP) 
  (set-process-status! aP interrupted) 
  (smc:switch aP *overlord* #f))

;;; General switching procedure. Puts the process somewhere and calls the 
;;; dispatcher. Special cases are reschedule and wait-duration (see below).
;;; Status is set by assuming that the parking place is not the scheduler!!!

(define (SMC:park aP waiting-place )
  (if (collection? waiting-place) (put! waiting-place aP))
     (set-process-status! aP interrupted) 
     (smc:switch aP *overlord* #f))
 
;;; Special cases of park:
;;; This puts the process into the ready list and then interrupts it.

(define (smc:reschedule aP args)
  (set-process-status! aP ready-to-run)
  (set-process-data! aP args)
  (put! (process-dispatcher-ready-list *overlord*) aP)
  (smc:switch aP *overlord* #f))


(define (smc:Wait-Duration aP time args)
  (set-process-status! aP waiting-for-event)
  (schedule-resume aP time args) ; to event list 
  (smc:switch aP *overlord* #f))
  
(define (smc:Wait-for-time aP time args)
  (set-process-status! aP waiting-for-event)
  (schedule-resume-abs aP time args) ; to event list 
  (smc:switch aP *overlord* #f))

;;====================Several MACROS=====================================
;MACROS

;;;syntatic sugar for the definition of the leading character in the play
;;;i.e. Processes 

;;; These macros produces a method or function) than executes asynchronously 
;;; It hides inside a process object that is used to control de activation
;;; of the method.  They should be calles before calling the start-simulation
;;; method. When they are call, they perform initializations and became
;;; suspended until given control by the process dispatcher.
;;; You can control the life of a body of async code by using the functions in
;;; the user interface: abandon, park, schedule, reschedule and interrupt.
;;; You should always use self to refer to a process from inside its definition.
;;; However observe that the call to the method returns the process object,
;;; a very useful feature whenever you need the process, for instance to
;;; interrupt it, or control it otherwise.

(define (process-async-definition definition body type name)
  `(,(if (null? type) 'define (symbol-concatenate 'define- type)) ,definition
     (let ((self (instantiate process 
                              :name (symbol-concatenate ',name '- ',(gensym)))))
       (block init
              (smc:schedule self #f)
              (call/cc (lambda(cont)
                         (set-process-continuation! self cont)
                         (return-from init self)))
              ,@body
              (smc:abandon)))))

(defmacro (define-async-method definition . body)
    (process-async-definition definition body 'method (car definition)))
  
(defmacro (define-async-function definition . body)
    (process-async-definition definition body '() (car definition)))


;;;=====================  PROCESS-DISPATCHERS  =================================

;;; The object that manages other processes. We keep the continuation of the 
;;; dispatcher and resume it. Thus it is a true coroutine. Whenever anybody 
;;; needs the dispatcher, it invokes resume on *overlord*. The dispatcher 
;;; follows continuations. The whole looks like a single very long
;;; dynamically constructed call to dispatch. Eventually, dispatch returns, 
;;; because the ready-list is empty. This returns to the place where the *first*
;;; activation of dispatch was made, all else forgotten in the passing of 
;;; continuations.

(define-class process-dispatcher Process 
     (( = ready-list :initializer (lambda()(instantiate queue)))))


;;; The main generic and method.

(define-generic (smc:dispatch (aK process-dispatcher)))

;;; The method just picks the first element of the ready-list 

(define-method (smc:dispatch (aK process-dispatcher))
   (let loop ()
     (let ((new (get! (process-dispatcher-ready-list aK))))
       (if (not new) #f 
           (begin
            (smc:switch aK new (process-data new))
            (loop))))))

;;==========================EVENT-DISPATCHERS================================
;;; a process-dispatcher that knowS all about events and time. 
;;; No user methods here
;;;
;;; Global-Continuation is the continuation that stops the simulation.
;;; clock is the simulation clock, time keeping mechanism
;;; event-number: keeps a tally of the number of events and is used to
;;;        identify them uniquely.
;;; Last-event is the last event that happened. It is there for you to check.
;;;     It is useful in, e.g. impatience treatment. See SimmerEvents and test
;;;     program number 11.
;;; LAP-size is the maximum size of the LAP
;;; LAP is the Pending event List (in spanish Lista Acontecimientos Pendientes)


(define-Class Event-Dispatcher process-dispatcher
        (( = global-return :initializer (lambda()#f))
         ( = clock :initializer (lambda()0))
         ( = event-number :initializer (lambda()0))
         ( = last-event :initializer (lambda()#f))
         ( = LAP-size :initializer (lambda()100))
         ( = LAP   :maybe-uninitialized)))

(define-method (initialize! (aEd event-dispatcher))
      (set-event-dispatcher-lap! aEd
             (instantiate heap
               :size (Event-dispatcher-LAP-size aEd))))

;;; dispatch-events is a function for efficiency  reasons. It is not foreseen
;;; that dispatch events should be speciaalized on the class of event dispatcher
;;; therefore there is not a substantial loss of generality in making this a 
;;; function.

;;; an event with priority <= *ignore-value* is ignored. 
;;; This is a way of removing events from the LAP. There are several cases
;;; The method make-happen has funny properties. 

;;; 1.- In make-happen the process-dispatcher is recurrently given control, 
;;; and eventually will find empty its process list. Then returns #f, and a 
;;; return is made to the smc:dispatch method of the event-dispatcher Then a 
;;; new call to aux-dispatch-events happens . 
;;; 2.-The event calls a procedure that returns. Then make-happen returns. 
;;;

(define-generic (smc:dispatch-events (unS Event-dispatcher)))
(define-method  (smc:dispatch-events (unS event-dispatcher))
   (let ((x (get! (Event-Dispatcher-LAP unS))))
     (if (Event? x) 
       (if (> (Event-priority x) *ignore-value*)
         (begin
          (set-Event-Dispatcher-clock! unS (Event-time x))
          (set-Event-Dispatcher-last-event! unS x)
          (if (Event-Dispatcher-trace unS) (display-object x))
          (make-happen x)))
       (error "No pending events at time " (Event-Dispatcher-clock unS)))))

           
;;; This the combined dispatcher acting as process and event dispatcher. Any 
;;; Eventually the call to dispatch (the only one!) returns because it
;;; finds the ready list empty, once the long and winding road through
;;; process continuations has ended with all processes are stoped,
;;; and then event scheduling takes over.


(define-method (smc:dispatch (aK event-dispatcher))
  (let loop ()
    (call-next-method) ; <=== make-happen returns here recurrently!!!
    (smc:dispatch-events aK)
    (loop)))



 
;;=============================================================================
;;; THE Global Master *overlord* controls everything: time, lifes and destinies
;;=============================================================================
;;; Init simulation initializes the *overlord*. If you provide the Lap size
;;; as the optional argument it is used. else the default size is used. 
;;; The user may also like to initialize *random* by doing 
;;; (set! *random (make-random)) insuring that the random number stream is
;;; the same across runs.

(define (init-simulation . Ev-size)
  (set! *overlord* 
        (if  (null? Ev-size)
          (instantiate event-dispatcher :name "*overlord*")
          (instantiate event-dispatcher :name "*overlord*" 
                       :LAP-size (car Ev-Size))))
    (set! *executing* *overlord*))

;;; finally, start the simulation. Store the global continuation to stop
;;; everything dead on its track!.

(define (start-simulation)
  (call/cc
   (lambda(cont)
     (set-Event-Dispatcher-global-return! *overlord* cont)
     (SMC:dispatch *overlord*)))
  'End!!!)

(define (stop) ((event-dispatcher-global-return *overlord*) 'END))



;;;==========================================================================
;;; handy functions that hide *executing*. Useful to control running Processes

(defmacro (Reschedule)`(SMC:Reschedule self #f))

(defmacro (wait-duration time . args)
     (if (null? args) 
       `(smc:wait-duration self ,time #f)
       `(smc:wait-duration self ,time ,@args)))

(defmacro (Wait-for-time time . args)
  (if (null? args)
    `(smc:wait-for-time self ,time #f)
    `(smc:wait-for-time self ,time ,@args)))

;;; and finally some user interface

(define abandon smc:abandon)
(define schedule smc:schedule)
(define interrupt smc:interrupt)
(define park smc:park)

;==========================================================================
