
; SimmerDrawing.scm
(require 'points)
;--------------------------------------------------------------------------
;; This is a minimal implementation of graphics, just enough to plot
;; graphic-stat-vars, plus some minimal frills and laces. Most of the code
;; should be portable. Use of mac specific drawing routines has been 
;; restricted to 
;;          1) move-to and line-to, that move the drawing point to a 
;;             given point an draw a line, respectively.
;;          2) the method physical-draw for the class graph-strings
;; There are two levels of drawing. The physical drawing, generic physical-draw
;; and the logical drawing implemented thru the generic picture-draw.
;; Physical drawing actually moves the pen in the screen. It takes place in the
;; screen from a description of the picture to be drawn in terms of lines and
;; moves (lineto and moveto). Drawing takes place as the picture is being
;; interpreted. It works by using the mac primitives.
;; It uses a plot area object to transform world coordinates into screen
;; coordinates. As long as you only use moveto and lineto, there is no need to
;; worry about those details. Just draw by using them in your own coordinates
;; but set-up a plot area to take care of the transformation. Read below to 
;; know more about setting up the plot area. 
;;
;; Logical drawing is an altogether different story. Drawing is performed by 
;; putting graphical objects in a drawing area, inside a  plot-area.  
;; add-go is the method that adds graphical objects to the plot area.
;;
;; No screen action takes place until you call "draw", a method on the plot-area
;; object. Then the whole bunch of things you have put into the plot
;; area is physically drawn. 

;; Since normally you want to draw in a window I have provided in the 
;; simmerWindows package a very simple window and draw package. Just call 
;; the function manage-window, giving as argument a plot-area
;; and a new window will be created, plot-area drawn and the window will be 
;; managed, meaning that you can move it, close it, etc. 
;; 
;; The normal procedure is to define a graphical object, composed of graphical
;; objects that are added to the plot-area using add-go. There are a few 
;; primitive objects I hacve defined that have defined picture-draw methods,
;; and therefore inmediately usable. Polygon, function, axis and graph-string 
;; are the elementary graphical objects. 

;;----------------------Class POINTS METHODS-----------------------------------


;; operations for points, giving inexact results

;;; adds two points
(define-generic (point-add (aP1 point) (aP2 point)))
(define-method (point-add (aP1 point) (aP2 point))
   (instantiate point :y (exact->inexact(+ (point-y ap1)(point-y ap2)))
                      :x (exact->inexact(+ (point-x ap1)(point-x ap2)))))

;;; subtracts two points
(define-generic (point-dif (aP1 point) (aP2 point)))
(define-method (point-dif (aP1 point) (aP2 point))
   (instantiate point :y (exact->inexact(- (point-y ap1)(point-y ap2)))
                      :x (exact->inexact(- (point-x ap1)(point-x ap2)))))

;;; componentwise division
(define-generic (point-ratio (aP1 point) (aP2 point)))
(define-method (point-ratio(aP1 point) (aP2 point))
   (instantiate point :y (exact->inexact(/ (point-y ap1)(point-y ap2)))
                      :x (exact->inexact(/ (point-x ap1)(point-x ap2)))))

;;; increase x
(define-generic (add-to-x (ap1 Point) incr))
(define-method (add-to-x (ap1 Point) incr)
      (instantiate point :y (point-y ap1) :x (+ (point-x ap1) incr)))

;;; increase y
(define-generic (add-to-y (ap1 Point) incr))
(define-method (add-to-y (ap1 Point) incr)
      (instantiate point :y (+ (point-y ap1) incr) :x (point-x ap1)))

;;-----------------------------BASIC DRAWING FRAMEWORK------------------------
;; The class plot-area plays the central role in drawing. One draws in a 
;; plot-area. A plot area is associated with a window through the port slot.
;; There are two coordenated sets in a plot-area: window and world coordenates.
;; window coordenates follow the Mac convention, with the positive y axis
;; runnig down from  in the North Western corner of the screen, the x axis as
;; usual. 
;;       Screen-top is a point with the screen coordenates of the NW corner,
;;       Screen-bottom has the screen coordenates of the SE corner.
;; World coordenates work as usual in Mathematics.
;;       World-top has the world coordenates of the NE corner of the world.
;;       World-bottom the same but for the SW corner.
;; The world->window transformation maps world coordinates to screen coordinates
;; mapping normal math coordinates into funny mac coordinates.
;; world->window is a closure that converts points.
;; Port should be associated with a port (a window).
;; cartoon is the drawing list. To draw,place the object in cartoon by using the
;; method add-go or the high level drawing routines.
;; Beware!. The actual drawing is done LIFO , the last object placed in cartoon
;; is the first to be drawn. This should be taken into account whenever using
;; sequence dependent drawing. (see picture-draw for axis objects) 
                           
(define-class plot-area object
       ((= screen-top) 
        (= screen-bottom)
        (= world-top)
        (= world-bottom)
        (= world->window :maybe-uninitialized)
        (= port :maybe-uninitialized)
        (= cartoon :initializer (lambda()'()))))

(define-method (initialize! (apl plot-area))
    (set-plot-area-world->window! apl (make-world->window apl))
    apl)

; Logical drawing level routine.  Puts things in cartoon to be drawn later on.
(define-generic (add-go (aPA plot-area) x))
(define-method (add-go (aPA plot-area) x)
     (set-plot-area-cartoon! aPa 
                             (cons x (plot-area-cartoon apa))))

; main drawing method. Draws in port anything that is in cartoon, by
; using the low level method physical-draw on the objects sitting there.
(define-generic (draw (aPA plot-area)))
(define-method (draw (aPA plot-area))
   (map (lambda(x)(physical-draw x apa))(plot-area-cartoon aPa)))
 
 
;---------------------------------------------------------------------------
;;; Auxiliary function
; creates a plot area with the given extreme point coordinates. Incrementing
; the window number and getting its placement in the screem from the list
; in *positions*

(define (create-plot-area max-value min-value tmax tmin namepa)
  (let
    ((incx (* 0.15 (- tmax tmin)))      ;leave 15% space
     (incy (* 0.15 (- max-value min-value)))) ; leave 15% space
    (instantiate plot-area 
        :screen-top (make-point 0 0)
        :screen-bottom (make-point 300 200)
        :world-top (make-point (+ tmax incx)(+ max-value incy)) 
        :world-bottom (make-point (- tmin incx)(- min-value incy))
        :port (make-window (window-rectangle (window-number)) namepa))))

;-----------------------------------------------------------------------------
;; generates the world window slot. No need for the user to fiddle with it.
(define-generic (make-world->window (aPa plot-area)))
(define-handy-method (make-world->window (aPa plot-area))
   (let*
     ((p11 (exact->inexact
            (/ (- (point-x screen-bottom) 
                  (point-x screen-top))
               (- (point-x world-top)
                  (point-x world-bottom)))))
      (p10 (- (point-x screen-top)
              (* p11 (point-x world-bottom))))
      (p22 (exact->inexact
            (/ (- (point-y screen-top) 
                  (point-y screen-bottom))
               (- (point-y world-top)
                  (point-y world-bottom)))))
      (p20 (- (point-y screen-bottom)
              (* p22 (point-y world-bottom)))))
     (lambda(x)
       (make-point (inexact->exact (round (+ p10 (* p11 (point-x x)))))
                   (inexact->exact (round (+ p20 (* p22 (point-y x)))))))))


;;--------------------The main drawing procedures-------------------------

(define (moveto aP aPa)
  (let
    ((a0 ((plot-area-world->window aPa) aP)))
    (mac#moveto (plot-area-port aPa) (point-x a0)(point-y a0))))

(define (lineto aP aPa)
  (let
    ((a0 ((plot-area-world->window aPa) aP)))
    (mac#lineto (plot-area-port aPa) (point-x a0)(point-y a0))))

;;----------------------Objects that can be drawn-------------------------

(define-class graphic-object object (): virtual)
(define-generic (physical-draw (aGO graphic-object)(apa Plot-Area)))
;;------------------------------------------------------------------------
;; Special classes of objects and their low level drawing procedures.
;; if you want to draw objects, you have to
;;    1) Define the class.
;;    2) Define physical-draw, using move-to line-to, or low level procedures
;;       that will make the program even less portable.

;; Lines given by the two end-points.
(define-class line graphic-object ((= initial-point)(= end-point)))
(define-method (physical-draw (aLine line)(aPa plot-area))
    (moveto (line-initial-point aLine) apa)
    (lineto (line-end-point aLine) apa))

;; Open polygons, given by the list of vertices. If you want a closed 
;; polygon make the las vertex the same as the initial one.
      
(define-class polygon graphic-object 
    ((= points :initializer (lambda()'()))))

(define-method (physical-draw (aP Polygon) (apa Plot-area))
   (let ((lista (polygon-points aP)))
     (if (pair? lista)
       (let loop ((x0 lista)(x1 (cdr lista)))
         (if (pair? x1) 
           (begin
            (moveto (car x0) apa)
            (lineto (car x1) apa)
            (loop (cdr x0)(cdr x1))))))))

;;------------------------------- Upper level drawing-----------------
;; The method picture-draw, draws by placing objects into cartoon for some
;; plot area. However for efficiency purposes you may use lower level
;; procedures. Again beware of the lack of portability.

(define-generic (picture-draw (ax graphic-object) apa))
 
(define-method (picture-draw (aP polygon) apa)
   (add-go apa aP))

;-----------------------------------------------------------------------------
;; draws and labels axis either horizontal (type=0) or vertical (type=1)
(define-class axis graphic-object
          ((= type)         ; 0 = horizontal 1 = vertical
           (= origin)       ;point giving the origin
           (= length)       ;length of the axis
           (= tick-separation) ; in world units
           (= tick-size :maybe-uninitialized)))  ;in world units


(define-handy-method (picture-draw (ax axis) apa)
    (let*
      ((num-ticks (round (exact->inexact (/ length tick-separation))))
       (end (if (zero? type) (add-to-x origin length)(add-to-y origin length)))
       (tick-value-init (if (zero? type)(point-x origin)(point-y origin)))
       (next-tick (make-point (if (zero? type) tick-separation 0)
                              (if (zero? type) 0 tick-separation)))
       (tick-line (make-point (if (zero? type) 0 tick-size)
                              (if (zero? type) tick-size 0))))
      (add-go apa (make-line origin end))
      (do
        ((i 0 (+ 1 i))
         (tick-value tick-value-init (+ tick-value tick-separation))
         (x origin (point-add x next-tick)))
        ((> i num-ticks) #t)
        (add-go apa 
                (make-graph-string 
                 (number->string (format (+ 0.000001 tick-value) 1))
                 (if (zero? type) 1 2)
                 (if (zero? type) 0 1) 9))
        (add-go apa (make-line x (point-dif x tick-line))))))


; Drawing Strings. Horizontal and vertical aligments should be specified.
; since the string is rooted at the last point moved-to (lend-to) the
; aligment is given with respect to that point.
; posible alignments are Left,Center,Right (0,1,2 )  for Horizontal and  
; Top,Center,Bottom  (0,1,2) for vertical alignment. To increase efficiency
; I use low level procedures in the physical-draw method.

(define-class graph-string graphic-object
    ((= text )
     (= alignment-H :initializer (lambda()'0))
     (= alignment-V :initializer (lambda()'2))
     (= size :initializer (lambda()9))))

(define-method (physical-draw (aS Graph-string)(apa Plot-area))
     (let*
       ((port (plot-area-port apa))
        (str  (graph-string-text aS))
        (height (graph-string-size aS))
        (width (mac#stringwidth port str)))
       (case
         (graph-string-alignment-H aS)
         ((1) (mac#move port (- (truncate (/ width 2))) 0))
         ((2) (mac#move port (- width) 0)))
       (case
         (graph-string-alignment-V aS)
         ((1) (mac#move port 0 (truncate (/ height 2))))
         ((0) (mac#move port 0 height))) 
       (mac#textsize port height)
       (mac#drawstring port str)))


              
