
(require 'collections)
(require 'slibhash)
(require 'stat-vars)


;;=================Interface to HASH TABLES from SLIB===================
;;Handling of files as hash tables. Files can be defined by the user by
;;instantiating the class, and optionally, giving the desired size.
;; Everything else is self explaining. Uses Slib.
;I do not like automatic size expansion, because, in simulation, it is a way of 
;delaying the unavoidable, and spending computer (and our time) waiting for
;a simulation that will eventually overflow memory. However, under user control
;you can use expand-file below!!. Remember that you can set the size when
;instantiating by doing (instantiate file :size <whatever you like>). The
; system defines the hash table to a prime number greater than the requested 
; size for n<20000. For greater values the size is not garantied to be a prime
; although it normally will be.

(define (prime-greater x)
  (let
    ((primos '(3 5 7 11 13 17 19 23 29 31 37 41 43 47 53 59 
               61 67 71 73 79 83 89 97 101 103 107 109 113 
               127 131 137 139 149))
     (raiz (truncate (sqrt x))))
    (define (factor? x lista)
      (and (pair? lista) 
           (<= (car lista) raiz)
           (not (= x (car lista)))
           (or (zero? (modulo x (car lista)))
               (factor? x (cdr lista)))))
    (define (return-prime x)
      (if (factor? x primos) (return-prime (+ x 2))
                                  x))
    (if (zero? (modulo x 2)) (return-prime (+ x 1))
                             (return-prime x))))


;---------------------------FILES--------------------------------------------
(define-class File keyed-Collection
              ((= size :initializer (lambda() 61))
               (= hTable :maybe-uninitialized)
               (= key :initializer (lambda() equal? ))))


(define-method (initialize! (aF File))
        (set-file-size! aF (prime-greater (file-size aF)))
        (set-File-hTable! aF (make-vector (file-size aF) '()))
        aF)

; The main acces functions put! and fread
(define-method (fread (aF file) x)
   (let ((nn (hash x (file-size aF))))
     (if (keyed-collection? (vector-ref (file-htable aF) nn))
       (fread (vector-ref (file-htable aF) nn) x)
       #f)))

(define-method (put! (aF file) arg)
  (let
    ((x (hash (object-key arg) (file-size aF))))
    (if (not (keyed-list? (vector-ref (file-htable aF) x)))
      (vector-set! (file-htable aF) x (instantiate keyed-list))) 
    (put! (vector-ref (file-htable aF) x) arg))) 

(define-method (remove-elt (aF file) key)
   (let
     ((x (hash key (file-size aF))))
     (remove-elt (vector-ref (file-htable aF) x) key)))


; this is stardard collection stuff deal with the full record
;arg shoud be an object of the type named-object. Puts arg in the file
;adressed by its name

(define-method (is-in? (aF File) arg)
    (fread aF (key arg)))

(define-method (make-get-elt (aF File))
  (let
    ((i (- (vector-length (file-Htable aF)) 1))
     (items '()))
    (define (find-next j)
      (if (< j 0) #f
         (let
           ((x (vector-ref (file-Htable aF) j)))
           (if (null? x) 
             (find-next (- j 1))
             (if (not (null? (keyed-list-data x))) 
               (begin
                (set! i (- j 1))
                (set! items (keyed-list-data x)))
               (find-next (- j 1)))))))

      (define (keyed-lista-elem)
        (if (null? items) (find-next i))
        (if (null? items) #f
            (let
              ((x (car items)))
              (set! items (cdr items))
              x)))
      (lambda() (keyed-lista-elem))))

;; expansion function. Increases by w or 50 (if you do not say how much)
;; it is a slow procedure that rehashes the full file. Beware!!

(define-method (expand-size (aF file) . w)
   (let
     ((x (instantiate file :size (+ (file-size aF) (if (pair? w)(car w) 50)))))
     (for-each-elt aF (lambda(z)(put! aF z)) x)))


;;------------------ STAT FILES--------------------------------------------

;;a file that collects timed statistics of the number of objects in file.
;; redefine the class (or redefine number in an initialize! method) if you 
;; want other types of statistics.

(define-class stat-file file 
     ((= number :inmutable :maybe-uninitialized)))

(define-method (initialize! (aSF stat-file))
   (call-next-method)
   (init-stats-gatherer! aSF
          (instantiate time-var :name (qualify-name aSF "number-in-file"))
          'number)
   aSF)
               
(define-method (put! (aF stat-File) datum)
       (call-next-method)
       (increment (stat-file-number aF) (current-time)))

(define-method (remove-elt (aF stat-file) key)
        (call-next-method)
        (decrement (stat-file-number aF) (current-time)))

;; this is provided since for stat-files is easy!!!. No so for regular files!.
(define-method (empty? (aF stat-file))
     (zero? (var-value (stat-file-number aF))))

(define-handy-method (display-object (aF stat-file) . port)
    (display-object number))
 