(require 'stat-vars)
(require 'simulation)
;;;
;;; ==========================Gates and CONDITIONS =============================
;;;  This is an implementation of semaphore coordination among processes.
;;;  Coordinating processes share acces to an object that has all stuff
;;;  required for the job.
;;;
;;  Defines the class "Gate" 
;;; A Gate is a waiting place for a single process waiting to cross the gate.
;;; A gate is always closed except when open by signalling it. The gate
;;; becomes closed inmediately after the process passes it.
;;;
;;; "Parking-place" is the place where the process will be put into storage,
;;; until the semaphore is set to pass
;;;  
;;;      wait-for :forces a wait the Gate (in the parking place)  
;;;      signal :Allows passing of the process on the parking place.
;;;      does not suspend the calling process. The passing process is scheduled
;;;      for execution


(define-class Gate Named-object
   ((= parking-place :initializer (lambda()#f))))

(define-generic (wait-for (aC Gate)))
(define-generic (signal (aC Gate)))

(define-method (wait-for (aTO Gate))
  (set-Gate-parking-place! aTO *executing*)
  (smc:interrupt *executing*))

(define-method (signal (aTO Gate))
  (let ((x (Gate-parking-place aTO)))
    (if x (begin
           (set-Gate-parking-place! aTO #f)
           (smc:schedule x #t)))))

;;; Defines the class "Condition". Generalizes Gate by providing a semaphore
;;; and a collection (a FIFO queue) to place waiting processes. 
;;;  A Condition is essentially a gate with a waiting placethat allows many
;;;  processes to wait in queue.
;;; "Parking-place" is the queue of processes waiting to enter the condition
;;; "Stop" is the condition semaphore. Test it if you want to know the state
;;; of the condition.
;;  
;;       wait-for :waits inside a condition(in the entry queue of the condition)
;;              if the semaphore is red.
;;       signal :Forces the semaphore to #f, allows passing of the first
;;              process on the entry queue. Since signal suspends the current
;;              process it gives a chance to the passing process to run. 
;;              To get another process to run you need to subclass signal.
;;       signal-all :Forces the semaphore to #f and allows passing of all
;;              processes in the entry queue. Since signal suspends the current
;;              process it gives a chance to all passing processes to run.
;;       reset  : forces the conditions semaphore to red i.e. puts it to #t
 

;; You can get additional conditions by replacing any collection for the FIFO 
;; queue. 
;; Statistics may be easily gathered by instantiating different types of
;; stat-things in the slot parking-place.


;;------------------------Conditions--------------------------------------

(define-class condition Gate
    ((= stop :initializer (lambda()#f))))


(define-method (initialize! (aC condition))
   (set-field-value!  aC
        (instantiate queue :name (qualify-name aC "condition-queue"))
        'parking-place)
   aC)

;-------------------------------------------------------------------------
;; A New generic for conditions

(define-generic (signal-all (aC Condition)))

;;; The inherited methods change the behavior of conditions in the way described
;;; above
;;;
;;; Wait-for tests the condition without trying to enter it, and therefore 
;;; without being caught in the entry queue if the semaphore is #t.

(define-handy-method (wait-for (aC condition))
     (if stop (smc:park *executing* parking-place))) 

;forces the semaphore of the condition to #f and allows entry of first process
;in queue
(define-handy-method (signal (aC Condition))
    (set-condition-stop! aC #f)
    (let
      ((x (get! (condition-parking-place aC))))
      (if x (smc:schedule x #t)))) 

;forces the semaphore of the condition to #f and allows entry of all processes
;in queue
(define-method (signal-all (aC Condition))
  (define (launch queue)
    (if (empty? queue) 
      #t
      (let
        ((x (get! queue)))
        (smc:schedule x #t)
        (launch queue))))
  (set-condition-stop! aC #f)
  (launch (condition-parking-place aC)))

; resets the condition to non-passing state. (i.e. resets stop to #t)
(define-method (reset (aC condition))
    (set-condition-stop! aC #t))

(define-method (display-object (aC Condition) . port)
   (display-object (condition-parking-place aC)))

;;=======================================================================
;;                            FACILITIES
;;=======================================================================

;; a facility is a container that has an assigned number of recources and 
;; assigns them among resource demanding processes. Therefore a facility can 
;; hold a maximum number of resource demanding processes.
;;
;; Use facilities to model a strech of code where only a given number of
;; processes could be working at one time. Facilities are a nice way of
;; modeling scarce resources. On entry each process grabs a number of 
;; units of the resource. Only max-number of units are available. Once they
;; are used up, any additional process must wait for resources to be freed
;; to be able to enter the restricted stretch of code.
;;
;;  The main methods defined for facilities are
;;       enter :enter the facility domain as another ocupant. Gain access to
;;              all variables in the shared object. If the condition does not
;;              allow passing you are directed to an entry queue where you
;;              wait until the condition is available.
;;       leave :the process leaves the facility
;;  You can use facilities with just those methods, but if you want to 
;; change the basic logic, you should 
;; subclass facilities and specialize the following methods, that determine
;; the behavior of the facilities on entry and leave 
;;
;;    get-In :Does the actual statistics management  whenever
;;              a process gets inside a facility
;;    get-out :The same, on leaving a facility
;;    test     :checks the facility and returns its status (#t or #f). 

;; by specializing those three methods you can get different behaviors. 
;; The default behavior is
;;
;;   get-in increases the number inside the condition by its argument
;;   get-out Same but now decreases the number inside by its argument
;;   test    Returns #t if the number inside plus the requested number
;;           is less or equal to max-number, one of the slots of the condition 
;;           setable by the user.
;;
;; therefore the default behavior is to allow inside the condition a number 
;; of processes as long as they do not require more than max-number resources,
;; and block the entering processes otherwise. Notice that if the front process
;; can not gain acces to the facility, all other procvesses are blocked even if 
;; the number of available resources would allow some of them to pass. Thus the
;; first process blocks even the considertion of processing the other processes.
;;
;; By changing test you can get other nice behaviors. In general, by
;; specializing all three methods you can get many conditions.

(define-class facility condition
              ((= number :inmutable :maybe-uninitialized)
               (= max-number :initializer (lambda()1))))
 
(define-method (initialize! (aF facility))
    (call-next-method)
    (init-stats-gatherer! aF
          (instantiate step-var :name (qualify-name aF "fac.stats"))
          'number)
    aF)

(define-generic (enter (aC Facility) . args))
(define-generic (leave (aC Facility) . args))
(define-generic (get-in (aC Facility) . args))
(define-generic (get-out (aC Facility) . args))
(define-generic (test (aC Facility) . args))

(define-handy-method (enter (AC Facility) . args)
  (if (null? args) (set! args '(1)))
  (if (apply test aC args)
    (apply get-in aC args)
    (begin
     (set-process-data! *executing* args)
     (smc:park *executing* parking-place))))

(define-handy-method (leave (aC Facility) . args)
   (apply get-out aC args)
   (let
     ((x (peek parking-place)))
     (if (and x (apply test aC (process-data x)))
       (begin
        (get! parking-place)
        (get-in aC (process-data x))
        (smc:schedule x #t)))))

(define-handy-method (get-in (aC Facility) . args)
       (apply increment number  (current-time) args))

(define-handy-method (get-out (aC Facility) . args)
       (apply decrement number (current-time) args))

(define-handy-method (test (aC Facility) . args)
       (<= (apply + (time-var-value number) args) max-number))

(define-method (display-object (aC Facility) . port)
    (displayLine "Facility " (facility-name aC))
    (call-next-method)
    (display-Object (facility-number aC)))