(require 'collections)
(require 'stat-vars)
;
;--------------------------------------------------------------------
;;Simple FIFO queues
;;===================================================================
; The code is borrowed from Slib and adapted to MeroonV3
; Elements in a queue are stored in a list.  The last pair in the list
; is stored in the queue type so that datums can be added in constant
; time.
; Two special generics are defined: 
;          queue-rear  same but at last element
;          queue-push  puts an element *in front* of the queue
; all other methods specialize those of the superclass
;;-------------------------------------------------------------------
(define-class queue collection
  ((= first-pair :initializer (lambda()'()))
   (= last-pair :initializer (lambda()'()))))

;--------------------------------------------------------------------
(define-generic (queue-rear (q Queue)))
(define-generic (queue-push! (q Queue) . arg))

(define-handy-method (empty? (q Queue))
        (null? first-pair))

(define-handy-method (peek (q Queue))
       (let 
         ((x first-pair))
         (if (null? x) #f
                       (car x))))

(define-handy-method (queue-rear (q Queue))
  (let 
    ((x last-pair))
    (if (null? x) #f
                  (car x))))

 
(define-handy-method (queue-push! (q Queue) . arg)
;put new element in front of the queue
       (let*
         ((datum (car arg))
          (old-first-pair first-pair)
          (new-first-pair (cons datum old-first-pair)))
         (set! first-pair new-first-pair)
         (if (null? old-first-pair)
           (set! last-pair  new-first-pair)))
       q)
 
;; specializing get! and put!

(define-handy-method (put! (q Queue) arg)
       (let 
         ((new-pair (cons arg '())))      
         (cond ((null? first-pair)
                (set! first-pair new-pair))
               (else
                (set-cdr! last-pair new-pair)))
         (set! last-pair new-pair))
       q)

(define-handy-method (get! (q Queue))
  (let ((x first-pair)) 
    (if (null? x) #f
                  (let ((first-cdr (cdr x)))
                    (set! first-pair first-cdr)
                    (if (null? first-cdr) (set! last-pair  '()))
                    (car x)))))

;; and this is the generator function constructor.

(define-handy-method (make-get-elt (as Queue))
  (let
    ((els (queue-first-pair as)))
    (lambda()
      (if (pair? els)
        (let
          ((x (car els)))
          (set! els (cdr els))
          x)
        #f))))

(define queue-pop! get!)

;;======================================================================
;; A class of FIFO queues instrumented to keep statistics of their ocupation.
;; Only the methods that add or remove elements should be specialized.
;; if you want different types of statistics just subclass stat-queue and 
; initialize the subclass with
;; the desired type of variable in slot :number. For instance
;; (instantiate gtime-var :name (qualify-name "graphic") in initialization
;; will collect graphical time data in the variable with name "...-graphic".

(define-class stat-queue Queue
        ((= number :inmutable :maybe-uninitialized)))

(define-method (initialize! (aSQ stat-queue))
    (init-stats-gatherer! aSQ 
       (instantiate step-var :name (qualify-name aSQ "stats"))
       'number)
    aSQ)
;-----------------------------------------------------------------------

(define-handy-method (queue-push! (q Stat-queue) . arg)
   (increment number (current-time))
   (call-next-method))

(define-handy-method (put! (q Stat-queue)  arg)
   (increment number (current-time))
   (call-next-method))
 

(define-handy-method (get! (q Stat-queue))
    (let
      ((x (call-next-method)))
      (if x (decrement number (current-time)))
      x))

(define-handy-method (display-object (q stat-Queue) . stream)
   (displayLine "Statistics for queue " name)
   (display-object number)) 