
(require 'named-objects)
(require 'points)
;
;-----------------------------------------------------------------
;; Povides several types of statistical variables. A statistical variable
;;remembers a value, and collects statistics of its values. You can also reset
;;the statistics, reinitializing them and thus forgetting about the past
;;sequence of (presumably) transient values.
;;CLASSES =========================================================

;;The simplest type counts the number of observations and remembers 
;;the max and min values among those observations.

(define-class VAR named-Object 
                   ((= accepting :initializer (lambda()#t))
                    (= value :initializer (lambda()0))
                    (= number :initializer (lambda()0))
                    (= max-value :initializer (lambda()Minus-Infinity))
                    (= min-value  :initializer (lambda()Plus-Infinity))))

(define-handy-method (display-object (aVar var) . stream)
       (displayLine "Var."  name
                    " Value:"  (format value 4)
                    " max-value: " (if (= max-value Minus-Infinity) "***" 
                                    (format max-value 4))
                    " min-value : " (if (= min-value  Plus-Infinity)  "***"
                                    (format min-value  4))))

(define-generic (increment (aVar Var) . args))
(define-generic (decrement (aVar Var) . args))
(define-generic (disc-increment (aVar Var) . args))
(define-generic (disc-decrement (aVar Var) . args))

(define-method (increment (aVar Var) . args)
      (take aVar (+ (var-value aVar) 1)))

(define-method (decrement (aVar Var) . args)
      (take aVar (- (var-value aVar) 1)))

(define-handy-method (take (aVar Var) y)
  (if accepting 
    (begin
     (set! value y)
     (set! number (+ 1 number))
     (if (> y max-value) (set! max-value y))
     (if (< y min-value) (set! min-value  y)))))

(define-handy-method (reset (aVar Var))
       (set! value 0)
       (set! number 0)
       (set! max-value Minus-Infinity)
       (set! min-value  Plus-Infinity))
;;----------------------------------------------------------------
;; A variable type that collects means and variances of its observations.
(define-class STAT-VAR var 
                       ((= mean :initializer (lambda()0))
                        (= variance :initializer (lambda()0))))

(define-handy-method (display-object (x stat-var) . stream)
        (call-next-method)
        (displayLine "N.Observ: " number "mean:" (format mean 4)
                     " Variance: " (format (sqrt variance) 4)))

(define-handy-method (reset (x stat-var))
        (call-next-method)
        (set! mean 0)
        (set! variance 0))

(define-handy-method (take (x stat-var) y)
   (if accepting
     (begin
      (call-next-method)
      (let
        ((oldmean mean)
         (coef (- 1.0 (/ 1.0 number))))
        (set! mean (+ (* coef mean) (/ y number)))
        (let
          ((incmean (- oldmean mean))
           (incvalue (- y mean)))
          (set! variance
                (+ (* coef (+ variance (* incmean incmean)))
                   (/ (* incvalue incvalue) number))))))))

;;-----------------------------------------------------------------------
;;A variable type that collects time averages and variances of the observations.
;; Therefore it effectively computes averages over sample paths
;; Each observation given by a point, either a continuity-point or a 
;; discontinuity-point

(define-class time-var stat-var 
                       ((= tstart :initializer (lambda()0))
                        (= tmean  :initializer (lambda()0))
                        (= tvariance :initializer (lambda()0))
                        (= time :initializer (lambda()0))))

(define-handy-method (reset (x time-var))
       (call-next-method)
       (set! time (current-time))
       (set! tstart time)               
       (set! Tvariance  0)
       (set! Tmean 0))


(define-handy-method (take (aTV time-var) punto)
  (define (sqr x)(* x x))
  (if accepting
    (let
      ((tactual (point-x punto)) ;extracts argument
       (oldvalue value)
       (valor (if (discontinuity-point? punto) 
                (discontinuity-point-yminus punto)
                (point-y punto))))
      (set! punto (point-y punto))   ;modifies the arguments to call up!!!!
      (call-next-method)
      (if  (> tactual tstart)    ;only if capturing
        (let ((coef (/ (- tactual time) (- tactual tstart)))
              (oldmean tmean))
          (set! tmean (+ (* (- 1 coef) tmean) 
                         (* coef (* 0.5 (+ oldvalue valor)))))
          (set! tvariance
                (+ (- (sqr tmean))
                   (* (- 1 coef) (+ tvariance (sqr oldmean)))
                   (* coef 
                      (+ (* 0.33333333333333 (sqr (- valor oldvalue))) 
                         (* valor oldvalue)))))))
      (set! time tactual))))


(define-handy-method (display-object (x time-var) . stream)
        (call-next-method)
        (displayLine 
         "T.Initial: " (format tstart 2)
         " T : " (format time 2) " mean[T] :" (format Tmean 4)
         " Sigma[T] : " (format (sqrt Tvariance) 4)))

;;;
;;; Functions to increment and decrement a general time variable
;;;

(define-method (increment (aVar time-Var) . args)
    (take aVar (make-point (car args) (+ (var-value aVar) 1) )))

(define-method (decrement (aVar Var) . args)
    (take aVar (make-point (car args)(- (var-value aVar) 1) )))

;; a general time variable that always processes discontinuous observations.
;; This is useful whenever you want to use stime-vars but do not want the 
;; graphical capabilities that come with them

(define-class step-var time-var ())

(define-method (take (aSV step-var) punto)
    (set! punto (make-discontinuity-point (point-x punto) 
                                          (point-y punto)
                                          (step-var-value aSV)))
    (call-next-method))
                     
