
;;==========================Objects to sample from============================
; This is the class of distributions, that is of object that can generate 
; samples from a probability distribution, and in general compute probabilities.
; All of them are derived from a basic object. There is an object for each type
; of distribution. 

(define-class distribution object () :virtual)
;sample returns the object as default behavior
(define-generic (sample (aD distribution)) aD)
(define-generic (distribution-function (aD distribution) x))
(define-generic (density-function (aD distribution) x))

;--------------------Rejection Method---------------------------------------
; as is, works for distributions concentrated between 0 and 1, and densities
; with maximulm value less than 1.
(define (rejection-method aD)
  (define (rejection-standard)
    (let
      ((rand  (*random*)))
      (if (>= (density aD rand) (*random*)) rand (rejection-standard))))
  (rejection-standard))

;----------------------Uniform----------------------------------------------
(define-class uniform distribution
   ((= a :initializer (lambda()0))
    (= b :initializer (lambda()1))))

(define-method (sample (aU uniform))
    (+ (uniform-a aU) (* (- (uniform-b aU)(uniform-a aU)) (*random*))))

(define-method (distribution-function (aU uniform) x)
   (cond ((< x (uniform-a aU)) 0)
         ((> x (uniform-b aU)) 1)
         (else (/(- x (uniform-a aU))(- (uniform-b aU)(uniform-a aU))))))

(define-method (density-function (aU uniform) x)
   (cond ((< x (uniform-a aU)) 0)
         ((> x (uniform-b aU)) 0)
         (else (/ 1. (- (uniform-b aU)(uniform-a aU))))))

;----------------------Discrete---------------------------------------------- 
(define-class discrete distribution 
     ((= value-list :initializer (lambda()'(0)))
      (= prob-list :initializer (lambda()'(1)))))

(define-method (sample (aD discrete))
  (let
    ((pcum (*random*)))
    (do
      ((x (discrete-prob-list aD)  (cdr x))
       (y (discrete-value-list aD) (cdr y))
       (sum 0 (+ sum (car x))))
      ((>= (+ sum (car x)) pcum) (car y)))))
                     

;----------------------Exponential-----------------------------
(define-class exponential distribution 
     ((= mean :initailizer (lambda()1))))

(define-method (sample (aE exponential))
   (- (* (exponential-mean aE) (##log (*random*)))))

(define-method (density-function (aE exponential) x)
   (* (exponential-mean aE) (exp (- (/ x (exponential-mean aE))))))

(define-method (distribution-function (aE exponential) x)
   (- 1 (exp (- (/ x (exponential-mean aE))))))

;----------------------Erlang-----------------------------------------
(define-class erlang distribution 
    ((= mean :initializer (lambda()1))
     (= n :initializer (lambda()1))))

(define-method (sample (aE Erlang))
  (do
    ((i (erlang-n aE) (- i 1))
     (x 0 (- x (##log(*random*)))))
    ((zero? i)(* x (/ (erlang-mean aE) (erlang-n aE))))))


;----------------------Normal----------------------------------------------
(define-class normal distribution 
   ((= mean :initializer (lambda()0))
    (= var :initializer (lambda()1))))

(define-method (sample (aN normal))
  (define (norm-inverse num)
      (let
        ((t (##sqrt (##log (/ 1 (* num num))))))
        (- t
           (/ (+ 2.30753 (* t 0.27061))
              ( + (* t (+ (* t 0.04481) 0.99229)) 1)))))
  (let 
      ((num (*random*)))
      (if (< num 0.5)  
        (+ (normal-mean aN) (* (normal-var aN) (norm-inverse num)))
        (- (normal-mean aN) (* (normal-var aN) (norm-inverse (- 1 num)))))))

(define-method (distribution-function (aN Normal) x)
    (define (auxnormal x)
      (let
        ((t (/ 1. (+ 1. (* 0.33267 x)))))
        (- 1 (* .3989422804014327 
                (exp (- (* 0.5 x x)))
                t
                (+ 0.4361836 (* t (+ -0.1201676 (* t 0.9372980))))))))
    (let (( z (/ (- x (normal-mean aN))(normal-var aN))))
      (if (< z 0) (- 1 (auxnormal (- z) )) (auxnormal z))))
 
(define-method (density-function (aN normal) x)
    (let ((z (/ (- x (normal-mean aN))(normal-var aN))))
      (/ (* .3989422804014327 (exp (- (* 0.5 z z))))(normal-var aN))))
   
;----------------------Log Normal-------------------------------------------
(define-class log-normal distribution 
              ((= mean :initializer (lambda()1))
               (= var :initializer (lambda()1))))

(define-method (sample (aL log-normal))
  (let*
    ((theta (/ (log-normal-var aL)(log-normal-mean aL)))
     (sigln (##sqrt (##log (+ 1 (* theta theta)))))
     (meanln (- (##log (log-normal-mean aL)) (* 0.5 (* sigln sigln )))))
    (##exp (sample (make-normal meanln sigln)))))


;----------------------Beta ----------------------------------------------

(define-class beta distribution
    ((= x0 :initializer (lambda()0))
     (= x1 :initializer (lambda()1))
     (= a :initializer (lambda()1))
     (= b :initializer (lambda()1))))

(define-method (density-function (aB beta) x)
   (let ((mid (+ (beta-a aB)(beta-b aB))))
     (/ (* (expt x (beta-a aB)) (expt (- 1 x) (beta-b aB)))
        (* (expt (/ (beta-a aB) mid) (beta-a aB))
           (expt (/ (beta-b aB) mid) (beta-b aB))))))
 
(define-method (sample (aB beta))
      (+ (* (- (beta-x1 aB)(beta-x0 aB)) (rejection-method aB))
         (beta-x0 aB)))

;--------------------Binomial--------------------------------------
(define-class binomial distribution
   ((= n :initializer (lambda()1))
    (= p :initializer (lambda()0))))

(define-method (sample (aB binomial))
  (do ((i 0 (+ i 1))
       (sum 0 (+ sum (if (< (*random*)(binomial-p aB)) 1 0))))
      ((> i (binomial-n aB)) sum)))

;--------------------Poisson---------------------------------------
(define-class poisson distribution
   ((= landa :initializer (lambda()1))))

(define-method (sample (aP poisson))
   (do ((i 0 (+ i 1)) 
        (sum 0 (+ sum (/ (- (log (*random*)))))))
       ((> sum (poisson-landa aP)) i)))
         
;--------------------Hypergeometric--------------------------------
(define-class hyper-geometric distribution
   ((= n :initializer (lambda()1))
    (= p :initializer (lambda()1))))

(define-method (sample (aH hyper-geometric))
   (do ((i 0 (+ i 1)) 
        (sum 0 (+ sum (if (< (*random*) (hyper-geometric-p aH)) 1 0))))
       ((> sum (hyper-geometric-n aH)) i)))

