; Dynamic Time Functions
;;;----------------------------------------------------------------------------
(require 'graphics)
(require 'time-functions)
;;;
;;; This module implements a calss of functions that are plotted as they 
;;; are computed. When the function is created, the system creates a window 
;;; and initializes it. In the window a frame is drawn  using as bounds the
;;; data supplied by the user.
;;;
;;; Every time a point is take'n by the function, the point is displayed in
;;; the window. Therefore you can see the evolution of the function as the
;;; simulation proceeds.
;;;
;;; Now this type of functions is intended for debugging purposes. Normally
;;; the user will design a simulation that computes and then displays the 
;;; results. Normally it is pointless to draw all results while they are being
;;; computed. This slows down the simulation and clubbers the screen.
;;;
;;; But sometimes you want to see what is happening as you go along. The 
;;; solution is turning some of the model time functions into to 
;;; dynamic-time-functions. Just this, will automatically create the windows
;;; and start the drawing process in real time.
;;;
;;; To this end, we provide several functions that scan an object and turn
;;; all functions appearing in its slots into dynamic-time-functions.
;;; Two kinds of procedures exist for this purpose. Shallow replacement that
;;; only replaces slots appearing at the first level of the object. The second 
;;; type of procedures is recursive procedures, that replace all appearences 
;;; of time-functions in any level of the object.
;;;
;;; Windows are created at initialization time for the object defining the
;;; traced functions.
;;;
;;; In dynamic graphs one should be careful, because the user can move the
;;; window in the middle of a redrawn and last positions can be juggled.
;;; we do not care about those details. The best way to use this feature is
;;; create the dynamic graphs., arrage the windows, and the run the simulation
;;; with the windows in place. Also be aware that whilst using dynamic
;;; graphs the interrupt does not work (who knows why!!). But I do not want to
;;; deal with those details. It all started as a simple exercise...
;;;----------------------------------------------------------------------------

;;; Implementation notes
;;; All there is in the implementation of this feature is the definition
;;; of the new class, and the redefinition of its the take method. In addition,
;;; the plot-area is now stored in the dynamic-time-function object.

(define-class dynamic-time-function time-function 
              (minx maxx mint maxt
              (= pantalla-nom :maybe-uninitialized)
              (= pantalla :maybe-uninitialized)))


(define-handy-method (initialize! (aDTF dynamic-time-function))
  (let ((framexy 
          (make-xy-frame (make-point mint minx)(make-point maxt maxx))))
    (set! pantalla (create-plot-area maxx minx maxt mint pantalla-nom))
    (picture-draw framexy pantalla)
    (manage-window pantalla)
    aDTF))

;;; It is very simple, but becomes a bit complicated because of 
;;; discontinuity points. Also we take precautions moving to teh 
;;; last point, just in case that the user moves the window and in
;;; redrawing it the system changes the location of the origin of the
;;; line.This is the reason behind the (seemingly) useless moveto's!!!

(define-handy-method (take (aDTF dynamic-time-function) ap)
   (define (jump-bottom-point x)
    (make-point (discontinuity-point-x x)(discontinuity-point-yminus x)))
   (if (last-point aDTF)
     (if (discontinuity-point? ap)
       (let ((x0 (last-point aDTF))
             (x1 (jump-bottom-point ap)))
         (add-go pantalla (make-line x0 x1))
         (moveto x0 pantalla)
         (lineto x1 pantalla)
         (add-go pantalla (make-line x1 ap))
         (moveto x1 pantalla)
         (lineto ap pantalla))
       (begin
        (add-go pantalla (make-line (last-point aDTF) ap))
        (moveto (last-point aDTF) pantalla)
        (lineto ap pantalla)))
     (moveto ap pantalla)) 
   (call-next-method))

;------------------------------------------------------------------------------
;User functions
;------------------------------------------------------------------------------
;; Replaces dynamic-time-functions for time functions in all levels  


(define (set-gdebugging! anO t0 ymin t1 ymax . fields)
  (let ((num (make-counter)))
    (define (replace-tf xx)
      (duplicate (xx dynamic-time-function)
                 :pantalla-nom (qualify-name anO (num))
                 :mint t0 :maxt t1 :minx ymin :maxx ymax))
    (define (final-name x)(string-append x "/"(number->string (num))))
    (define (inner-debug-vector x)
      (for i 0 (- (vector-length x) 1)
           (cond ((sure-is-an-object? (<- x i))
                  (if (time-function? (<- x i))
                    (-> x i (replace-tf (<- x i))) 
                    (inner-debug-vector (<- x i))))
                 ((vector? (<- x i))(inner-debug-vector (<- x i)))
                 ((list? (<- x i)) (inner-debug-list (<- x i))))))
    (define (inner-debug-list x)
      (let loop ((xx x)) 
        (if (pair? xx)
          (begin
           (cond ((sure-is-an-object? (car xx))
                  (if (time-function? (car xx))
                    (set-car! xx (replace-tf (car xx)))
                    (inner-debug-vector (car xx))))
                 ((vector? (car xx))(inner-debug-vector (car xx)))
                 ((list? (car xx)) (inner-debug-list (car xx))))
           (loop (cdr xx))))))
      (for-each 
        (lambda(x)(cond ((vector? (field-value anO x))
                         (inner-debug-vector (field-value anO x)))
                        ((list?   (field-value anO x))
                         (inner-debug-list (field-value anO x)))))
        fields)))


(define (reset-gdebugging! anO . fields)
    (define (inner-debug-vector x)
      (for i 0 (- (vector-length x) 1)
           (cond ((object? (<- x i))
                  (if (time-function? (<- x i))
                    (-> x i (duplicate ((<- x i) time-function)))
                    (inner-debug-vector (<- x i))))
                 ((vector? (<- x i))(inner-debug-vector (<- x i)))
                 ((list? (<- x i)) (inner-debug-list (<- x i))))))
    (define (inner-debug-list x)
      (let loop ((xx x)) 
        (if (pair? xx)
          (begin
           (cond ((object? (car x))
                  (if (time-function? (car x))
                    (set-car! xx
                              (duplicate ((car x) time-function)))
                    (inner-debug-vector (car x))))
                  ((vector? (car x))(inner-debug-vector (car x)))
                  ((list? (car x)) (inner-debug-list (car x))))
           (loop (cdr xx))))))
     (for-each 
      (lambda(x)(cond ((vector? (field-value anO x))
                        (inner-debug-vector (field-value anO x)))
                       ((list?   (field-value anO x))
                        (inner-debug-list (field-value anO x)))))
      fields))

; estas funciones hacen solo shallow debugging. del primer nivel del objeto.

(define (shallow-set-gdebugging! x t0 ymin t1 ymax)
  (for i 1 (- (vector-length x) 1) 
          (if (time-function? (<- x i))
            (-> x i (duplicate ((<- x i) dynamic-time-function)
                               :pantalla-nom (get-name x)
                               :pantalla #f
                               :mint t0 :maxt t1
                               :minx ymin :maxx ymax)))))

(define (shallow-reset-gdebugging! x)
    (for i 1 (- (vector-length x) 1) 
           (if (dynamic-time-function? (<- x i))
             (-> x i (duplicate ((<- x i) time-function))))))

                 
;;;=======================Ejemplo de Uso ============================
; GRAPHICAL DEBUGGING 
; use it as follows

;(define pepe (instantiate gtime-var))
;(shallow-set-gdebugging! pepe 0 0 10 10)
;(for i 0 10 (take pepe (make-point i (* (*random*) 10))))
;(shallow-reset-gdebugging! pepe)
;(for i 11 20 (take pepe (make-point i (* (*random*) 10))))
;(display-object pepe)

