
;;;==============================EVENTS====================================
;;; There are several types of events: Normal events and recurrent events. 
;;; Normal events execute a function (a method) once. Recurrent events
;;; reschedule copies of themselves to be executed some future time.

;----------------EVENT CONTROL-----------------------------------------------
;;; You have control over the events thru the posibility of inspecting the LAP,
;;; searching for a particular event with a given number. You can then inspect
;;; event and act accordingly. also the last occurred event is always available 
;;; for your inspection.

;---------------CONTINUOUS SIMULATION via RECURRENT EVENTS--------------------
;;; You can perform continuous time simulation, i.e. performing some methods 
;;; every time step. This is done thru the use of recurrent events. A recurrent 
;;; event is an event that happens repeatedly with a time between events given 
;;; by the user. 
;;; The typical use of a recurrent event is performing chores that do not depend
;;; on time coordination of processes. It allows a method to be executed 
;;; repeatedly, without entering the scheduler (not using continuations), thus
;;; avoiding (most of) the time consuming dispatching.
;;; Also, you can specify different intervals, therefore having several heart
;;; beats, one for each relevant object. By using event priorities you can make
;;; sure that the order of executing is the desired one, whenever there are ties
;;; in the event execution time. 
;;; Finally you can specify a delay time, that delays the start of the recurrent
;;; event making it a delayed recurrent event.

;;; The following functions are the most useful to control events.
;;;
;;;     create-event  loads an event given by : time message data
;;;     create-event-with-priority loads an event with the given priority 
;;;     create-recurrent-event creates a repeating, or recurrent, event.
;;;     load-Events loads a list of events for execution.
;;;     get-last-event returns the last event that happened
;;;     ignore-event the given event (already scheduled)
;;;                  is ignored by the dispatching process.
;;;
;;; To find out more on control functions, see below.
;;;
;;; In defining an event, "message" is the function that will be executed, and 
;;; it will be applied to the arguments in "data". Normally the first argument 
;;; in data will be the object to whom a message is sent. 
;
; --------------------------NORMAL EVENTS---------------------------------
;;; the stuff to signal time.
;;; Time: The event will happen when the clock signals this time.
;;; Message: This is the procedure that will be invoked (applied).
;;; Data: The Message will be applied to the contents of this slot 
;;; Priority: Tie breaking whenever two events have the same time.

(define-class Event Object (Time Message Data Priority))

;;; This is the main generic that performs an event.

(define-generic (make-happen (anEv Event)))
 
;;; An event with priority <= to *ignore-value* is ignored!!!!. This is a
;;; way of removing events from the LAP. 

(define *ignore-value* -1)

;;; define a method to order events. They are not ordered by their keys!!!!.
;;; They are ordered first by time and then by priority.

(define-method (object<? (aE1 event)(aE2 event))
   (let ((x (event-time aE1))
         (y (event-time aE2)))
   (if (= x y)
     (< (event-priority aE1)(event-priority aE2))
     (< x y))))

;;; a Method to print the event information.
(define-handy-method (display-object (anEv Event) . stream)
    (DisplayList  " Ev:" (format (user-event-time anEV)  3) " Pr:" priority 
                  " Mess:" message " Data:")
    (for-each show data)
    (newline))

;;; the method that makes normal events happen
                                   
(define-handy-method (make-happen (anEv Event))
   (apply message data))

;-------------------------RECURRENT EVENTS----------------------------------
;;; Recurrent events allow pure event oriented simulation. A recurrent event 
;;; schedules a chain of events at intervals of time. Since events may 
;;; specify the performing of a method, then any sequence of methods can be 
;;; executed at a sequence of time intervals.
;;; There are two main classes of Recurrent events:
;;; Fixed step size: The time interval to the next event is deterministic
;;;       and always the same. Therefore events are evenly spaced.
;;; Random step size: The time interval is defined by a probability distribution
;;; You can create other subclasses, for instance executing different methods
;;; at different time intervals, which is akin to constructing a pure event
;;; oriented simulation model.

(define-class Recurrent-event Event
        ((= recurrence-time :initializer (lambda()0))))

;;; generic that allows the association of recurrent events with an object. You
;;; can define a method for this generic that creates the recurrrent events
;;; you want. Then you call it and forget about everything else. t0 is the
;;; time where the first event of a recurrent chain will be scheduled.

(define-generic (start-object-process (anObj object) t0 . args))

;;; auxiliary function that pulls functionality for all subclasses. It clones
;;; the event, modifies its time of executions and schedules it.

 (define (aux-make-happen anEv next-time)
   (let ((x (clone anEv)))
     (set-Recurrent-event-time! x  (time-round  (+ (current-time) next-time)))
     (put! (event-dispatcher-LAP *overlord*) x)
     (apply (recurrent-event-message anEv)(Recurrent-event-data anEv))))
 
;;; The main method for recurrent event happening.

(define-method (make-happen (anEv Recurrent-event))
  (aux-make-happen anEv (recurrent-event-recurrence-time anEv)))

;;;------------------------RANDOM RECURRENT EVENTS -------------------------
;;; This allows for variable intervals between events given by a function in
;;; the slot recurrence-time. Notice the extra set of parenthesis around the
;;; recurrence-time slot!
;;; The event expects a function in this slot, that will be evaluated to 
;;; determine the interval to the next event.

(define-class Random-Recurrent-event Recurrent-Event ())

;;; And here is the main method
(define-method (make-happen (anEv Random-Recurrent-event))
   (aux-make-happen anEv ((recurrent-event-recurrence-time anEv))))

;;;--------------------support functions for event management ----------------
;;; The following functions help in creating and managing events.

; internal function that enters events. Event time is rounded to prevent silly
; effects.

(define (event time message arg pr)
  (let ((x (make-event (time-round time) message arg pr)))
    (if (put! (Event-dispatcher-LAP *overlord*) x) x
        (error "No room in Pending Events List when loading events"))))

;; creates absolute events.
(define (create-abs-event time message . arg)
  (event time message arg 0))

; creates events relative to current time.
(define (create-event time message . arg)
  (event (time-from-now time) message arg 0))

;; event dispatcher scheduling relative to current-time
(define (schedule-resume aP time . args)
      (apply create-event time smc:resume aP args))

;; absolute dispatcher scheduling. Use with a lot of care!!!
(define (schedule-resume-abs aP time . args)
  (apply create-abs-event time smc:resume aP args))

; creates events from a list of tuples of the form (time object message . data)
(define (load-events event-list)
  (for-each (lambda(x)(apply create-event x)) event-list))

; creates events allowing the assignment of priorities
(define (create-event-with-priority time message pr . args)
  (event (time-from-now time) message args pr))

; Creates recurrent events
; of the form <initial delay> <Priority> <time-between-events> <message> <data>
; all events derived from a recurrent event have the same number

(define (create-recurrent-event time Priority Rec-time message . arg)  
  (let ((x (make-recurrent-event 
             (time-round (time-from-now time)) message arg  priority rec-time)))
    (if (put! (event-dispatcher-LAP *overlord*) x) x
        (error "No room in Pending Events List when loading events"))))

;=============================================================================
;;;  IMPATIENCE
;=============================================================================
;;; The following procedures form the skeleton of the impatience mechanism. 
;;; It is very elementary and not completely foolproof. 
;;; You can always interrupt a wait *not in the event list*  by scheduling an
;;; impatience event that starts some alternative path. However you need
;;; acces to the events in the LAP to find event that was supossed to
;;; terminate the wait in a normal way. This can be done by
;;; 1.- Keeping track of the event in a variable, using schedule-resume to 
;;;     generate the impatience and put it into the LAP. Then start the wait.
;;; 2.- Once the reawakening of the process happens, you should test
;;;     for the event (with eq?) to see if it is a normal or impatience exit. 
;;;     An impatience exit will be signalled by the impatience event. Then 
;;;     perform the impatience code.
;;; 3.- If normal then you should remove the impatience event from the LAP. You
;;;     can do it by using your stored (pointer to) event and ignore-event.
;;; 
;;; This can be used for all waits for conditions, but not for waits for a 
;;; given duration. This is because for an scheduled wait that has an associated
;;; impatience event, any one happening requires removing the other. This
;;; in turn requires getting the event number of the normal event, and 
;;; removing it if impatience occurs. This can be done by the user, but then 
;;; you can not use wait-duration because it does not get thorugh the event
;;; numbers. A new macro culd be defined along the lines in the exemple 11.
;;; It is left as a future extension, if needed...

;;; The function ignore-event changes an event for the given event and marks 
;;; its priority to *ignore-value*. When *overlord* finds a marked 
;;; priority ignores the event. Thus ignore-event efectively suppreses an event.

(define (ignore-event anEv)
  (set-event-priority! anEv *ignore-value*))

(define (get-last-event)(event-dispatcher-last-event *overlord*))