; (require 'time-functions)
; histograms and methods to compute and draw them


;; A histogram is a class that knows how to draw itself in the right way.
;; Now, a histogram is created from a time-function or a stime.var. 
;; In the first case it computes the histogram of values. In the second case 
;; computes the histogram of time percentajes. This second histogram only has a
;; meaning when the function is a step function, hence the restriction.


(define-class histogram graphic-object
    ((= name :initializer (lambda()""))
     (= cells :maybe-uninitialized)
     (= cell-freq :maybe-uninitialized)
     (= num-cells :initializer (lambda()3))))

(define-method (initialize! (aH histogram))
    (call-next-method)
    (set-histogram-cells! aH (make-vector (histogram-num-cells aH)))
    (set-histogram-cell-freq! aH (make-vector (histogram-num-cells aH) 0))
    aH)


;;; This generic can be specialized to any kind of object. It computes the
;;; histogram.

(define-generic (compute-Histogram (aH histogram)(aF object))
    (error "Histograms can not be computed for class " 
           (class-name (object>class aF))))

(define-method (compute-Histogram (aH histogram) (aF time-function))
   (with-access aH (histogram cells cell-freq num-cells)
     (let* ((xmin (- (min-value aF) 0.0000001))
            (xmax (+ (max-value aF) 0.0000001))
            (step (exact->inexact (/ (- xmax xmin) (- num-cells 1)))))
       (for i 0 (- num-cells 1) (-> cells i (+ xmin (* i step))))     
       (let loop ((puntos (queue-first-pair (time-function-points aF))))
         (if (pair? puntos)
           (let ((iplace (inexact->exact 
                           (floor (/ (- (point-y (car puntos)) xmin)
                                     step)))))
             (-> cell-freq iplace (+ (<- cell-freq iplace) 1))
             (loop (cdr puntos)))))
       aH)))

;;; When given a stime-var compute-histogram computes the histogram of time 
;;; percentajes of a function, i.e the total % of time a function has the 
;;; value y.


(define-method (compute-Histogram (aH histogram)(aF stime-var))
   (with-access aH (histogram cells cell-freq num-cells)
     (let* ((aFun (stime-var-data aF))
            (xmin (- (min-value aFun) 0.0000001))
            (xmax (+ (max-value aFun) 0.0000001))
            (range (- xmax xmin))
            (rangex (- (max-time aFun) (min-time aFun)))
            (step (exact->inexact (/ range (- num-cells 1)))))
       (for i 0 (- num-cells 1)  (-> cells i (+ xmin (* i step))))       
       (let loop ((puntos (queue-first-pair (time-function-points aFun))))
         (if (pair? (cdr puntos))
           (let ((iplace (inexact->exact
                           (floor (/ (- (point-y (car puntos)) xmin)
                                     step)))))
             (-> cell-freq iplace 
                 (+ (<- cell-freq iplace) 
                    (- (point-x (cadr puntos))
                       (point-x (car puntos)))))
             (loop (cdr puntos)))))
       (for i 0 (- num-cells 1) (-> cell-freq i (/ (<- cell-freq i) rangex)))
       aH)))


;;; displaying a histogram in a graphical way. If you want the values, display
;;; cells and cell-freq side by side.

(define-handy-method (picture-draw (aH histogram) aPa)
   (define (plot-rectangle x0 y0 x1 y1)
     (add-go aPa (make-line (make-point x0 y0)(make-point x0 y1)))
     (add-go aPa (make-line (make-point x0 y1)(make-point x1 y1)))
     (add-go aPa (make-line (make-point x1 y1)(make-point x1 y0)))
     (add-go aPa (make-line (make-point x1 y0)(make-point x0 y0))))
   (for i 0 (- (histogram-num-cells aH) 2)
        (plot-rectangle (<- cells i) 0 (<- cells (+ i 1))(<- cell-freq i))))

(define-handy-method (display-object (aH histogram) . port)
   (let* ((max-val (apply max (vector->list cell-freq)))
          (framexy (make-xy-frame 
                     (make-point (<- cells 0) 0)
                     (make-point (<- cells (- num-cells 1)) 
                                 (* 1.1 max-val ))))
          (pantalla (create-plot-area (* 1.1 max-val) 0 
                                      (<- cells (- num-cells 1))
                                      (<- cells 0) 
                                      name)))
     (picture-draw framexy pantalla)
     (picture-draw aH pantalla)
     (manage-window pantalla)))

;;=============================================================================
;; ---function to create and draw the histogram of the values of a function ---
;;=============================================================================

(define (display-histogram-for aF numcells name)
  (let* ((incells (+ numcells 1))
         (histo (instantiate histogram :num-cells incells :name name )))
   (display-object (compute-histogram histo aF ))
   histo))

              
