;;SimerTimeFunctions
(require 'points)
(require 'graphics)
(require 'queues)
;

;; first define functions to be subclassed. Little else can be done at this 
; level of generality. Even drawing a function is an exercise in ingenuity. 
; Therefore leave the thing very general and expect the user to provide 
; methods to subclass the generics.

(define-class function graphic-object () :virtual)

;a method to check for well-definedness
(define-generic (well-defined? (aTF function)) #t)

; the fundamental method that computes a value for the function
(define-generic(Valor (aF Function) z))

; the integral definite
(define-generic (Definite-integral(aT function) a b))

; combination of two functions by the operator f.
(define-generic (Operate-functions (x function)(y function) f))



;;===========================================================================
; This is a rather complete implementation of time-functions, mainly for
; continuous system simulation. Graph vars store their graphic data in a
; time function. You add points by creating a point and using the method take 
; to send it to the (graph-var) or time-function. To send a discontinuity point ; to a graph-var, create it and do a take. 

; a time function is represented by a queue of points. It is a queue so that 
; additions with increasing t are added at the end of the list. The queue is in
; the variable points. You can add normnal points or discontinuity points.
; It  handles jump discotinuities in a nice way. 
; All methods know and test, about discontinuity points. 
; The function is defined for the whole real axis. Below the first point is 
; given by the linear extrapolation based on the first two points. Beyond the
; x value of the last point it is constant and equal to the last value of y.

;;=================================================

(define-class time-Function function
   ((= header :initializer (lambda()""))
    (= points :initializer (lambda()(instantiate queue)))
    (= previous :initializer (lambda()'()))
    (= current :initializer (lambda()'()))
    (= current-value :initializer (lambda()Plus-infinity))))

         
;This is a method to add points to a time-function. 
(define-method (take (aTF time-function) x)
    (put! (time-function-points aTF) x))

; an auxiliary function that allows the definition of a general function by
; giving the points, one after the other, in the argument list

(define (define-function . x0)
  (let ((f (instantiate time-function)))
    (let loop ((x x0))
      (cond ((pair? x)
             (cond ((= (length (car x)) 3) 
                    (take f (apply make-discontinuity-point (car x))))
                   ((= (length (car x)) 2)
                    (take f (apply make-point (car x))))
                   (else (error "Dato invalido en definicion de funcion")))
             (loop (cdr x))))
            f)))

(define-method (well-defined? (aTF time-function))
   (peek (time-function-points aTF)))

; This is an auxiliary function that constructs a time-function from two vectors
; the x vectyor (with the values of t) and the f vector (with the values of the
; function)

(define (define-time-function x f)
  (let ((x0 (instantiate time-function)))
    (do ((i 0 (+ i 1)))
        ((= i (vector-length x)) x0)
      (take x0 (make-point (vector-ref x i)(vector-ref f i))))))

;; methods to retrieve the value of a function for a given argument. The
;; returned value depends on the type of the data, characterized by the
;; type of function in the slot data. It the function slot step is false, 
;; then an interpolation is performed and the value returned. 
;; Else the step fucntion value is returned.

(define-generic(position (aF time-function) x))
(define-method (Position (aF time-function) x)
  (define (listSearInner lista0 lista1)
    (if (pair? lista0)
      (if (<= (point-x (car lista0)) x) 
        (listSearInner (cdr lista0) lista0)
        (begin
          (set-time-function-current! aF lista0)
          (set-time-function-previous! aF lista1)
          (set-time-function-current-value! aF x)
          (list lista0 lista1)))
      (error "invalid status of a function " x)))
  (if (< x (time-function-current-value aF))
    (listSearInner 
     (queue-first-pair (time-function-points aF))
     (queue-first-pair (time-function-points aF)))
    (listSearInner 
     (time-function-current aF) (time-function-previous aF)))) 


; You should be careful when using valor, specially in discontinuity points
; since giving a real as z can play havoc with your values, due to the 
; representation problem for real numbers in a binary machine.!!!


(define-method (valor (aF Time-Function) z)
   ;; Interpola linealmente entre pj y pi, dos puntos teniendo en cuenta
   ;; las posibles discontinuidades
   (define (interpolate pj pi)
     (if (= z (point-x pj)) 
       (point-y pj)
       (let ((y1 (point-y pi))
             (y2 (if (discontinuity-point? pj) (discontinuity-point-yminus pj)
                     (point-y pj)))
             (x1 (point-x pi)))
         (+ y1 (* (- z x1)  (/(- y2 y1) (-(point-x pj) x1)))))))
   ;;;
   ;;; Busca un intervalo donde interpolar
   (if (>= z (max-time aF)) 
     (point-y (queue-rear (time-function-points aF)))
     (let* ((x0 (Position aF z))
            (xj (car x0))
            (xi (cadr x0)))
       (if (eq? xi xj) (if (discontinuity-point? (car xi)) 
                         (discontinuity-point-yminus (car xi))
                         (interpolate (cadr xj)(car xj)))
           (interpolate (car xj)(car xi))))))

;auxiliary function to read the last point of a time function
(define (first-point ATF)(peek (time-function-points aTF)))
(define (last-point ATF) (queue-rear (time-function-points aTF)))

; some auxiliary methods to find the min and max time value of a time-function,
; ie the end points of the definition interval. THis is because you can remove 
; points at both ends. (its a queue...!)

(define (limit-at xx) 
  (if (discontinuity-point? xx)(discontinuity-point-yminus xx) (point-y xx))) 

(define (min-time aF)
  (let ((x0 (first-point aF)))
    (if x0 (point-x x0) plus-infinity)))

(define (max-time aF)
   (let ((x0 (last-point aF)))
     (if x0 (point-x x0) minus-infinity)))

(define (extreme-value aF operator initial)
  (let loop ((xx (queue-first-pair (time-function-points aF)))
             (extreme initial))
  (if (pair? xx)
    (let ((y (point-y (car xx)))
          (ymin (limit-at (car xx))))
      (loop (cdr xx)(operator extreme y ymin)))
    extreme)))
  
(define (min-value aF) (extreme-value aF min plus-infinity))
(define (max-value aF) (extreme-value aF max minus-infinity))

;;; and auxiliary function to find all the min and max time/value of a 
;;; time-function, uses multiple return

(define (bounds aF)
  (let loop ((xx (queue-first-pair (time-function-points aF)))
             (miny plus-infinity)
             (maxy minus-infinity)
             (minx plus-infinity)
             (maxx minus-infinity))
  (if (pair? xx)
    (let ((x    (point-x (car xx)))
          (y    (point-y (car xx)))
          (ymin (limit-at (car xx))))
      (loop (cdr xx)(min miny y ymin)(max maxy y ymin)(min minx x)(max maxx x)))
    (return-list minx miny maxx maxy))))

;------------------------------------------------------------------------------
;; all given points in a step-time-function are discontinuity points. 
;; a step function is continuous from the right. Outside the interval of 
;; definition it is constant and equal to the y of the first and las points
;; repectively.

(define-class step-time-function time-function ())

; Adds a point to the step function. You supply a point and it is converted 
; and filled with the adequate value for y

(define-method (take (aSTF step-time-function) x)
  (set! x (->discontinuity-point x))
  (set-discontinuity-point-yminus! x 
      (if (pair? (queue-first-pair(time-function-points aSTF)))
        (discontinuity-point-y (queue-rear(time-function-points aSTF)))
        0))
  (call-next-method))

(define (define-step-time-function x f)
  (let ((x0 (instantiate step-time-function)))
    (do ((i 0 (+ i 1)))
        ((= i (vector-length x)) x0)
      (take x0 (make-point (vector-ref x i)(vector-ref f i))))))

;;----------------------------------------------------------------------------
;; this is to draw the 2D-graph of a single variable function, including
;; jump discontinuiities 



(define-method (picture-draw (af Time-function) apa) 
  (define (jump-bottom-point x)
    (make-point (discontinuity-point-x x)(discontinuity-point-yminus x)))
  (define (draw-rest list0)
    (if (pair? (cdr list0))
      (let ((x (car list0))
            (y (cadr list0)))
        (if (discontinuity-point? y)
          (begin
            (add-go apa (make-line x (jump-bottom-point y)))
            (add-go apa (make-line (jump-bottom-point y) y)))
          (add-go apa (make-line x y)))
        (draw-rest (cdr list0)))
      #t))
  (draw-rest (queue-first-pair (Time-function-points af))))

(define-handy-method (display-object (aF Time-function) . args)
  (multiple-let ((x0 min-val x1 max-val)(bounds aF)) 
    (if (and (< min-val max-val)(< x0 x1))
      (let* ((framexy 
               (make-xy-frame (make-point x0 min-val)(make-point x1 max-val)))
             (name (if (pair? args)(car args) header))
             (pantalla (create-plot-area max-val min-val x1 x0 name)))
        (picture-draw framexy pantalla)
        (picture-draw aF pantalla)
        (manage-window pantalla)))))

;crea una funcion que cuando se le da un nombre pinta la lista de variables
;proporcionadas, incluyendo aTF y lista.

(define-method (display-object-list (aTF Time-function) . lista )
   ;;initializations. Reset the world_>window map!!!!
  (let* ((list (cons aTF lista))
         (theBounds (map bounds list)))
    (let ((t0   (apply min (map car theBounds)))
          (mmin (apply min (map cadr theBounds)))
          (t1   (apply max (map caddr theBounds)))
          (mmax (apply max (map cadddr theBounds)))
          (name (time-function-header aTF)))
      (if (and (> t1 t0)(> mmax mmin))
        (let
         ((pantalla (create-plot-area mmax mmin t1 t0 name))
          (framexy (make-xy-frame (make-point t0 mmin) (make-point t1 mmax))))
         (picture-draw framexy pantalla)
         (for-each (lambda(x)(picture-draw x pantalla)) list)
         (manage-window pantalla))))))

;-----------------------------------------------------------------------------
; it integrates for the range between min-value and max-value. Therefore it
; does not consider the values of the function outside this domain!!!!!

(define-handy-method (Definite-integral (aT time-function) a b)
  (define (inner-integral suma lista0)
    (let loop ((sum suma)
               (lista lista0))
      ;(displayLine sum lista)
      (if (and (pair? (cdr lista))(<= (point-x (cadr lista)) b)) 
        (loop (+ sum (* (+ (point-y (car lista))
                           (if (discontinuity-point? (cadr lista))
                             (discontinuity-point-yminus (cadr lista))
                             (point-y (cadr lista))))
                        (- (point-x (cadr lista))
                           (point-x (car lista)))
                        0.5))
              (cdr lista))
        (if (pair? (cdr lista))
          (+ sum (* 0.5 
                    (+ (valor aT b)(point-y (car lista)))
                    (- b (point-x (car lista)))))
          sum))))
  (if (<= a (min-time aT)) 
      (inner-integral 
        (* (-(min-time aT) a)(valor aT a))
        (queue-first-pair (time-function-points aT)))
      (inner-integral 
        0 
        (cons (make-point a (valor aT a))(car (position aT a)))))) 
 

; sampling from any function
(define-generic (sample-function (aF function) x0 x1 sampling-interval)
  (let ((result-function (instantiate time-function)))
   (do ((x x0 (+ x sampling-interval)))
       ((> x x1) result-function)
     (take result-function (make-point x (valor aF x))))))

;;==============================================================================
;      OPERATIONS BETWEEN FUNCTIONS
;-------------------------------------------------------------------------------
; Operates between time-functions accounting for the presence of discontinuities
; Knows when a discontinuity point should be replaced by a point, 
; because the jump has been cancelled. 


(define-method (Operate-functions (x Time-function)(y time-function) funcion)
 (define (primer-valor x)(if (pair? x) (point-x (car x)) plus-infinity))
 (let ((resultado (instantiate time-function)))
    ;Discontinuity removal
   (define (puntofinal x y0 y1)
      (if (= y0 y1) 
        (make-point x y0)
        (make-discontinuity-point x y0 y1)))
   
   ; Simplest case: no discontinuities
   (define (punto x y1 y2)
     (make-point x (funcion y1 y2)))
   
   ; Case with one discontinuity. Here z is a number.
   (define (punto0 x w1 z)
     (let ((x0 (funcion (discontinuity-point-y w1) z))
           (x1 (funcion (discontinuity-point-yminus w1) z)))
       (puntofinal x x0 x1)))
   ; This is needed because function mey not be conmutative!!
   (define (punto0c x w1 z)
     (let ((x0 (funcion z (discontinuity-point-y w1)))
           (x1 (funcion z (discontinuity-point-yminus w1))))
       (puntofinal x x0 x1)))
 
   ;Two discontinuities
   (define (punto1 x w1 w2)
     (let ((x0 (funcion (discontinuity-point-y w1)
                        (discontinuity-point-y w2)))
           (x1 (funcion (discontinuity-point-yminus w1)
                        (discontinuity-point-yminus w2))))
       (puntofinal x x0 x1)))
   
   (define (process-point minimo x val invertir)
     (if (discontinuity-point? x)
       (if invertir 
         (take resultado (punto0c minimo x val))
         (take resultado (punto0 minimo x val)))
       (if invertir
         (take resultado (punto minimo val (point-y x)))
         (take resultado (punto minimo (point-y x) val)) )))
   
   ; The central procedure that operates on functions. Lots of cases to be 
   ; considered
   ; First: which function provides the knot
   ; Second: what happens when both functions provide knots.
   ; There is a chance of removal of discontinuities depending on the operation
   
   (define (doOperation  a b)
     (let* ((a0 (primer-valor a)) 
            (b0 (primer-valor b)) 
            (minimo (min a0 b0)))
       (if (= minimo plus-infinity) '()
           (let  ((puntoa (if (pair? a) (car a)))
                  (puntob (if (pair? b) (car b))))
             (cond 
               ((and (= minimo a0 b0)  ; common x in both knots
                  (cond 
                    ((and (discontinuity-point? puntoa)
                          (discontinuity-point? puntob))
                     (take resultado (punto1 minimo puntoa puntob)))
                    ((discontinuity-point? puntoa)
                     (take resultado (punto0 minimo puntoa (point-y puntob))))
                    ((discontinuity-point? puntob) 
                     (take resultado (punto0c minimo puntob (point-y puntoa))))
                    (else 
                     (take resultado (punto minimo (point-y puntoa)
                                                   (point-y puntob)))))
                     (doOperation (cdr a)(cdr b))))
               ((= minimo a0) 
                (process-point minimo puntoa (valor y minimo) #f) 
                (doOperation (cdr a) b))
               (else          
                (process-point minimo puntob (valor x minimo) #t)
                (doOperation a (cdr b))))))))
     (doOperation 
      (queue-first-pair (time-function-points x))
      (queue-first-pair (time-function-points y)))
    resultado))

(define (sum-functions x y) (Operate-functions x y +))

;;============================================================================
;; some very elemntary methods to deal with real functions given by a formula
;; of Scheme function

(define-class real-function time-function 
     ((= procedure :maybe-uninitialized)))

(define-method (valor (aRF real-function) t)
     ((real-function-procedure aRF) t))

(define-method (well-defined? (aRF real-function))
   (field-defined? aRF 'procedure))


