;;; Uncomment the following 4 lines if you are not using a compiled version
;(require 'distributions)
;(require 'channels)
;(require 'priority-queues)
;(require 'procesos)
;(require 'graph-vars)

;;; This is a rather comprehensive implementation of a n-th server priority
;;; queue. If you want to see a trace of what is going on, just turn on the
;;; trace slot for the service-station object. To do that, uncomment the 
;;; call to trace, in the simulation code.
;;; In this exemple I use a channel to hold the incoming items -acting as 
;;; as messages. Each server has an async method executing concurrently with 
;;; others. Whenever the server work is over, it requests a new item from
;;; the channel. It becomes suspended, and waits in queue, until a new
;;; item arrives. All queues are priority queues, and therefore all waiting
;;; things, items or servers, get their rights in the order of priority.

;;; I create three objects: items, servers and service-stations. Items are 
;;; arriving customers. Service time may depend on the combination of an item 
;;; and a server. In the present implementation actual service time is the
;;; quotient between a standard procesing time - a propertie of the item- and
;;; a performance -a propertie of the server, assumed the same for all 
;;; apperances of the server. However, more complicated forms may be defined
;;; in the method serve, that has as  discriminating variables both the server 
;;; and the item.
;;;
;;; Arrival is made a method so that you can specialize it for subclasses of 
;;; items. This choice requires a template of an item supplied as argument. The
;;; argument is not used, other than in method selection, done by the system.  
;;; The needed instances of item are generated by the arrival method, in a loop 
;;; delayed by the interarrival time. Also the method gets as arguments the 
;;; various distributions of the problem, arrival, service and priority.
;;;
;;; Servers are generated by the method load-servers and put to work by  their
;;; method get-and-go-to-work. This last method calls serve, a method
;;; with two discriminating variables item and server, that does the actual 
;;; processing.
;;;
;;; When a server gets an item from the channel it becomes the value of its 
;;; customer slot, put there by the method take. This allows the cross use
;;; of items and servers in determining the properties of service time.
;;;
;;; finally service-station is the actual cahnnel, but it also acts as a 
;;; placeholder, collecting statistics, anchoring traces, etc.
;;;--------------------------------------------------------------------------
;;;                       Customers

(define-class item object
      ((= numero :initializer (lambda()(numera)))
       (= priority :maybe-uninitialized)
       (= birth-time :initializer (lambda()(current-time)))
       (= arrival :initializer (lambda()(current-time)))
       (= standard-service-time :maybe-uninitialized)))

(define-method (object-key (aIt item))
   (item-number aIt))

(define-method (object<? (a item)(b item))
  (cond
    ((< (item-priority a)(item-priority b)) #t)
    ((= (item-priority a)(item-priority b))(< (item-arrival a)(item-arrival b)))
    (else #f)))

(define-generic (arrivals (aI item) 
                      Dist-Arrivals  Dist-service Dist-Priorities cola))

(define-async-method (arrivals (aI Item) 
                     Dist-Arrivals  Dist-service Dist-Priorities cola)
  (let loop ()
    (let ((x (instantiate  item
         :birth-time (current-time)
         :standard-service-time (sample dist-Service) 
         :priority (sample dist-Priorities))))
      (if (named-object-trace cola)
        (displayLine (formatf (current-time) 6 1) "Arrival of :" 
                     (item-numero x) " Prior=" (item-priority x)))
      (set-item-arrival! x (current-time))
      (take cola x)      
      (wait-duration (sample dist-arrivals ))
      (loop))))
;---------------------------------------------------------------------------
;;;                        Servers
;;; Here I am using a different approach. Now each server has an async method
;;; describing its performance. Previous models hace assign an async method to
;;; a customer. The two approaches are complementary. This one is better when
;;; the server has a complex service routine. For exemple, when patrolling.
;;; I use a polyfield to collect statistics of the different priority classes.

(define-class server named-object 
              ((= performance :initializer (lambda()1))
               (= customer :maybe-uninitialized)))


;;; this is the processing method for a server
(define-generic (get-and-do-work (aS server) cola))
                     
(define-async-method (get-and-do-work (aS server) cola)
   (let loop ()
     ;the server is needed in give for priority determination in case of wait 
     (set-server-customer! aS (give cola aS)) 
     (serve aS (server-customer aS) cola)
     (loop)))


;;; The actual process of an item in a server
;;; a multiple method that allows processing to depend on both the server class
;;; and the item class. It is waitfor because the caller has to wait for the
;;; method to return before doing anything else.


(define-generic (serve (aS server)(aI item) cola))

(define-waitfor-method (serve (aS Server)(aI item) cola)
  (if (named-object-trace cola)
    (displayLine (server-name aS)
               "Starting service " (formatf (current-time) 6 1)
               (item-numero aI) (item-priority aI)))
  (wait-duration (/ (sample (item-standard-service-time aI)) 
                    (Server-performance aS)))
  (if (named-object-trace cola)
    (displayLine "Ending service "  (formatf (current-time) 6 1) 
               (item-numero aI)(item-priority aI)))
  (collect-stats cola aI))

;;;-----------------------------------------------------------------------
;;;                   The service system
;;;-----------------------------------------------------------------------

(define-class service-station channel
    ((= number-of-servers :maybe-uninitialized :inmutable)
     (= number-of-priorities :maybe-uninitialized :inmutable)
     (* prior-stats :initializer 
        (lambda(j)(instantiate time-var 
                     :name  (string-append "Pr-" (number->string j)))))))

(define-handy-method (initialize! (aSE service-station))
 (call-next-method)
 (set! message-list 
   (instantiate stat-priority-queue  :name 'System
                :number (instantiate sio-var :name "System-cola")))
  (initialize-field-value! aSE (service-station-prior-stats-length  aSE)        
                           'number-of-priorities)
  aSE)

;;; This loads the number of servers, creates the servers and initializes
;;; their performances

(define-generic (load-Servers (aSS service-station) n performances))
(define-method (load-Servers (aSS service-station) n performances)
   (initialize-field-value! aSS n 'number-of-servers)
   (for i 0 (- n 1) 
        (get-and-do-work 
          (instantiate server 
                       :performance (<- performances i)
                       :name (string-append "Server-" (number->string i)))
          aSS)))

;;;
;;; This is also a multi method to allow for cross dependencies.
;;;
(define-generic(collect-stats (aSS service-station) (aI Item))) 
(define-method (collect-stats (aSS service-station) (aI Item))
   (take (service-station-prior-stats cola (item-priority aI))
         (make-point (current-time)
                     (- (current-time)(item-birth-time aI)))))

;;----------------------------------------------------------------------------
; modelling a single queue with priorities. We use the technique of marking the
;items to get the distribution of the waiting time in system. We keep the 
;statistic gathering variables in a global vector.

;;;===========================================================================
;;;                              Simulate
;;;===========================================================================

(init-simulation)

(define numera (make-counter))
(define cola (instantiate service-station  :name 'Service 
                          :prior-stats-length 2 ))
(define cliente (instantiate Item ))
(load-servers cola 2 '#(1 1))
(set! *random* (make-random))
;(trace cola)   ; uncomment this to see a trace of arrivals and services
(arrivals cliente 
         (make-exponential 10) 
         (make-exponential 8)
         (make-discrete '(0 1) '(0.5 0.5))
         cola)
           
(create-event 400 stop)
(start-simulation)
(for i 0 (- (service-station-prior-stats-length cola) 1)
     (display-object (service-station-prior-stats cola i)))
(display-object cola)


;;=========================================================================
;;                       Priority servers
;;-------------------------------------------------------------------------
;;; To see the power of the approach, now I am going to model the case where
;;; servers have priorities, and that an incoming customer is served by the
;;; highest priority free server.  Of course the average results for the full
;;; system are exactly the same as before!!!.
;;; The trick is keeping processes ordered by priority in the channel. THis is
;;; easy by passing to the give method the server, having the priority slot.
;;; The ordering of processes is inherited from their data slots therefore
;;; since channels places its second argument (if any) in the data slot you
;;; just pass the priority setting object
;;;
;;; This is what is required:
;;;
;;; First define the new type of servers

(define-class priority-server server
    ((= priority :initializer (lambda()0))))

(define-method (object-key (aPS priority-server))
               (priority-server-priority aPS))

;;; and the method to order them. 

(define-method (object<? (x priority-server)(y priority-server))
   (< (priority-server-priority x)(priority-server-priority y)))


;;; now subclass the service station to have a priority queue as the queue 
;;; where processes (items) waiting for a message are held. Since servers
;;; wait there when idle, thats all there is to it.

(define-class priority-service-station service-station ())

(define-method (initialize! (aPSS priority-service-station))
   (call-next-method)
   (set-priority-service-station-queue! aPSS
       (instantiate stat-priority-queue 
                    :number (instantiate stime-var :name "servers")))
   aPSS)

;;; finally you need this method for initialization

(define-method (load-Servers (aSS priority-service-station) n performances)
   (initialize-field-value! aSS n 'number-of-servers)
   (for i 0 (- n 1) 
        (get-and-do-work 
          (instantiate priority-server 
                       :performance (<- performances i)
                       :name (string-append "PServer-" (number->string i))
                       :priority i)
          aSS)))

;;; thats all. Here is the main with some small changes to adapt it

;;--------define the system ---------------------------------------------
;; Here,init-simulation must preced everything since clock must be defined 
;; when initializing items.

(init-simulation)
(define numera (make-counter))
(define cola (instantiate priority-service-station  :prior-stats-length 2 ))
(define cliente (instantiate Item ))

;;---------Simulate
(load-servers cola 2 '#(1 1))
(set! *random* (make-random))
(arrivals cliente 
         (make-exponential 10) 
         (make-exponential 8)
         (make-discrete '(0 1) '(0.5 0.5))
         cola)
(untrace get-and-do-work object-key give get!)
(trace-off *overlord*)
(trace-off cola)   ; uncomment this to see a trace of arrivals and services        
(create-event 400 stop)
(start-simulation)
(for i 0 (- (service-station-prior-stats-length cola) 1)
     (display-object (service-station-prior-stats cola i)))
(display-object cola)



