;;; Uncomment the following line if you are not using a compiled version
;(require 'channels)

;;; This is an implementation of wait interruption.It is based on program 6 
;;; The implementation of impatience, or any other interrupted wait, is 
;;; accomplished by
;;;
;;; 1) Each event is assigned a number. We capture the number of an 
;;;    event upon creation.
;;; 2) On resumption, the case on hand is determined by examining the number
;;;    of the event. If equal to the stored one, it is the scheduled event. 
;;;    If something other, the wait has been interrupted.  Then the normal
;;;    event is removed from the LAP.
;;; 3) To interrupt a method you should have acces to the process. To do that, 
;;; you either capture the result of its initial call or assign self to a 
;;; global variable, for the interruptible method. Then you can perform an 
;;; smc:resume on it 
;;;
;;;    It is assumed that somebody issues a smc:resume call that resumes the
;;;    async procedure waiting for an event.
;;;
;;;---------------------------------------------------------------------------
(define numero #f)

;;; This is a macro that can be borrowed to define interuptible waits and 
;;; that performs the above steps 1 and 2.
;;; body is of the form {:normal expression} {:interrupt expression}
;;; The macro schedules a wait for a given time, and if interrupted erases
;;; the pending event . Otherwise it executes the indicated user code for 
;;; the two possible ways of ending the wait: interrupted or normal

(defmacro (Interruptible-wait-duration time . body)
  ; a form to construct an association list from the body
  (define (construct-pair-list tira)
    (if (null? tira) '()
        (cons (list (car tira) (cadr tira)) 
              (construct-pair-list (cddr tira)))))
  (let ((nnev (gensym))
        (pairs (construct-pair-list body)))
    (let ((normal-body (assq ':normal    pairs))
          (int-body    (assq ':interrupt pairs))) 
    `(let ((,nnev  (schedule-resume self ,time #f)))     ; to event list 
       (smc:interrupt self)
       ; if you are here it may be because of impatience. Test it.
       (if (and (event? (get-last-event))
                (eq?  (get-last-event) ,nnev))
         ,(if normal-body (cadr normal-body) #f)
         (begin ; Interrupted. Remove normal event.
          (ignore-event ,nnev)             
          ,(if int-body (cadr int-body) #f)))))))          

;;----------------Definitions from exemple 6 ----------------------------------
(define-class service-center channel 
      ((= system :maybe-uninitialized)
       (= dist-arrival :initializer (lambda()#f))))
  
(define-method (initialize! (aSC service-center))
    (call-next-method)
    (set-service-center-system! aSC
      (instantiate stat-queue :name (qualify-name aSC 'in-system)))
    aSC)

(define-generic (main (aSc service-center)))

(define-class server-item object 
      (( = dist-service :initializer (lambda()#f))))

(define-class customer object 
              ((= number :maybe-uninitialized)
               (= server-assigned :initializer (lambda()#f))))


(define-generic (service (aC customer) aservice-center))

;; the main program that generates items
(define-async-method (main (aSC service-center))
   (let loop ()
     (service (instantiate customer :number (numero)) aSC) 
     (wait-duration (sample (service-center-dist-arrival aSC)))
     (loop)))

;;----------------------New definitions---------------------------------------
;; This only works for a one server queue. We capture the waiting process in the
;; slot "in-service". Number-interrupted keeps statistics on interruptions.

(define-class interruptible-center service-center 
   ((= in-service :initializer (lambda()#f))
    (= number-interrupted :maybe-uninitialized)))

(define-method (initialize! (aIC interruptible-center))
    (call-next-method)
    (init-stats-gatherer! aIC
         (instantiate step-var :name (qualify-name aIC 'num-interr))
         'number-interrupted)
    aIC)

;;  the life story of an item

(define-async-method (service (aC customer) aservice-center)
  ; printing to trace execution
  (if (interruptible-center-trace aservice-center) 
    (displayLine "Entrado cliente " (customer-number aC) 
                 " a las " (current-time)))
  ; the full system
  (put! (interruptible-center-system aservice-center) aC)   ;store the customer 
  (set-customer-server-assigned! aC (give aservice-center ))    ;get server
  ; More printing
  (if (interruptible-center-trace aservice-center) 
    (displayLine "Inicio servicio " (customer-number aC) 
                 " a las " (current-time)))
  (set-interruptible-center-in-service! aservice-center self) ;capture process
  ; Now wait but expect interrupts!!
  (interruptible-wait-duration                                ;now wait
    (sample (server-item-dist-service (customer-server-assigned aC)))
    :normal (if (interruptible-center-trace aservice-center) 
              (displayLine "Salida "(customer-number aC) 
                           " Normal " (current-time)))
    :interrupt (begin
                (if (interruptible-center-trace aservice-center) 
                  (displayLine "Interrumpido" (customer-number aC) 
                               (current-time)))
                (increment
                  (interruptible-center-number-interrupted aservice-center)
                  (current-time)))) 
  (set-interruptible-center-in-service! aservice-center #f) ;nobody in service
  (take aservice-center (customer-server-assigned aC)) ;give back the server
  (get! (service-center-system aservice-center)))      ;remove the customer 

(define-async-function (interruptions aSC dist-interrupts)
   (let loop ((x (sample dist-interrupts)))
     (if (interruptible-center-in-service aSC)
       (smc:schedule (interruptible-center-in-service aSC) #f))
     (wait-duration x)
     (loop (sample dist-interrupts))))
      
(let(( unservice (instantiate interruptible-center
                              :name 'service
                              :dist-arrival (make-exponential 10)))) 
  (init-simulation)
  (trace-off unservice)
  (set! numero (make-counter))
  (set! *random* (make-random))
  (trace-off *overlord*)
  (create-event 1000 stop)
  (take unservice (instantiate server-item :dist-service (make-exponential 8)))
  (main unservice)
  (interruptions unservice (make-exponential 5))
  (start-simulation)
  (display-object unservice)
  (newline)
  (display-object (interruptible-center-number-interrupted unservice))
  (newline)
  (display-object (service-center-system unservice)))

