;;; Event Oriented simulation in the classical way.
;;;
;;; A queue simulated by using the pure event-oriented approach. We define two 
;;; type of events arrival and end-service. We provide handlers for the events 
;;; that are called by them. Responsibility of handlers is arranging the state 
;;; of the system so that after the event it is as desired.
;;;
;;; I am collecting too many statistics, specially in graphical form. It shows
;;; the power of Simmer, but also the danger of misusage. To avoid graphs, that
;;; just clutter everything when time is large, we display stime-vars by 
;;; typecasting them into time-vars. We keep the stime-vars as data collectors!.

;;; It is always much better to collect data  in the most general form and then 
;;; adjust the display-object method to your needs. This is so, because you do 
;;; not need to repeat the simulation to get other statsitics. To just change 
;;; the view of the situation just change and run display-object again.

;;;---------------------------------------------------------------------------
;;; Defining the classes of objects involved: service-system and customer
;;;---------------------------------------------------------------------------

(define-class service-system named-object
              ((= number-of-servers :initializer (lambda()1))
               (= service-dist :maybe-uninitialized)
               (= arrival-dist :maybe-uninitialized)
               (= busy-servers :initializer (lambda()(instantiate stime-var)))
               (= service-array :maybe-uninitialized)
               (= system :initializer (lambda()(instantiate stime-var)))
               (= queue :initializer (lambda()(instantiate stat-queue)))
               (= wait-stats :initializer 
                  (lambda()(instantiate time-function)))))

(define-handy-method (initialize! (aSS service-system))
   (set-time-var-name! busy-servers (qualify-name aSS 'busy-servers))
   (set-time-var-name! system (qualify-name aSS 'number-in-system))
   (set-stat-queue-name! queue (qualify-name aSS 'queue))
   (set-stat-queue-number! queue 
         (instantiate time-var :name (qualify-name aSS 'number-in-queue)))
   (initialize-field-value! aSS (make-vector number-of-servers #f) 
                            'service-array)
   aSS)

(define-class customer named-object
              ((= birth-time :initializer (lambda()0))
               (= server-assigned :maybe-uninitialized)
               (= wait-time :initializer (lambda()0))))

(define-handy-method (initialize! (aC customer))
   (set! birth-time (current-time))
   aC)

;;;-------------------------------------------------------------------------
;;; the simulation methods
;;;-------------------------------------------------------------------------

(define-generic (grab-a-server (aS service-system) aC))
(define-generic (arrival-handler (aS service-system)))
(define-generic (end-service-handler (aS service-system) aC))
                     
(define-handy-method (grab-a-server (aS service-system) aC) 
  (increment busy-servers (current-time))
  (do ((i 0 (+ 1 i)))
      ((not (<- service-array i) )
       (-> service-array i aC)
       (set-customer-server-assigned! aC i)))           
  (create-event (sample service-dist) end-service-handler aS aC))


(define-handy-method (arrival-handler (aS service-system))
   (let ((aC (instantiate customer)))
    ;put in system
    (increment system (current-time))
    (set-customer-wait-time! aC (current-time))
    ;if server available schedule service else wait in queue
    (if (< (var-value busy-servers) number-of-servers)
      (grab-a-server aS aC)
      (put! queue aC))
    ; schedule next arrival
    (create-event (sample arrival-dist)  arrival-handler aS)))

(define-handy-method (end-service-handler (aS service-system) aC)
    (-> service-array (customer-server-assigned aC) #f)
    (decrement busy-servers (current-time))
    (decrement system (current-time))
    (take wait-stats (make-point (current-time) 
                                 (- (current-time)(customer-wait-time aC))))
    (if (not (empty? queue))
      (grab-a-server aS (get! queue))))

;;;
;;; ------------------ Displaying Statistics -------------------------------
;;;
;;; Beware because typecast changes the class forever and therefore you
;;; need to draw th histograms before the typecasting is done, or else it wont 
;;; find the data slot. Also future actions will find the classes changed.To
;;; avoid all this is much better to restore the class after using typecast.

(define-handy-method (display-object (aS service-system) . port)
   (displayLine "--------------------RESULTS------------------------------")
   (displayLine "Service Center " (get-name aS))
   (displayLine "---------------------------------------------------------")
   (display-histogram-for busy-servers 2 "Number-of-busy-servers")
   (display-histogram-for wait-stats 10 "Waiting-time-Histogram")
   (display-histogram-for system 7 "Number-in-system-Histogram")
   (display-object (typecast busy-servers 'time-var))
   (typecast busy-servers 'stime-var)
   (newline)
   (display-object (typecast system 'time-var))
   (typecast system 'stime-var)
   (newline)
   (display-object queue)
   (newline))

;;;---------------------------------------------------------------------------
;;;                          THE SIMULATION
;;;---------------------------------------------------------------------------

(define station (instantiate service-system
                             :name 'Service-Center
                             :number-of-servers 2
                             :arrival-dist (make-exponential 10)
                             :service-dist (make-exponential 16)))
(let ((t0 (runtime)))
 (init-simulation)
 (set! *random* (make-random))
 (create-event 0 arrival-handler station)
 (trace-off *overlord*)
 (create-event 1000 stop)
 (start-simulation)
 (displayLine "Elapsed runtime " (- (runtime) t0)))

(display-object station)


   
      

